# $Id: handlers.py,v 1.6 2013/02/28 00:49:55 irees Exp $
'''File handlers.'''

import time
import math
import os
import signal
import struct
import json
import sys
import cPickle as pickle

# For file writing
import tempfile

# EMEN2 imports
# Provide a dummy handler if EMEN2 cannot be imported.
class DummyHandler(object):
    @classmethod
    def register(cls, *args, **kwargs):
        def f(o):
            return o
        return f
  
try:
    import emen2.db.handlers
    BinaryHandler = emen2.db.handlers.BinaryHandler
except:
    BinaryHandler = DummyHandler
    

# EMAN2 can only be imported in the main thread.
try:
    import EMAN2
    # We need to steal these handlers back from EMAN2...
    signal.signal(2, signal.SIG_DFL)
    signal.signal(15, signal.SIG_DFL)
except ImportError:
    EMAN2 = None


class EMDataBuilder(object):
    '''Helper class to build tiles and thumbnails for EMAN2-readable files.
    
    Ex:
    builder = EMDataBuilder()
    tile = builder.build("test.dm3", "test.dm3.tile")

    There is a kludge to write out some additional scaled images:
    
    copyout = {
        128: "thumb.jpg",
        512: "small.jpg"
    }
    builder = EMDataBuilder()
    tile = builder.build("test.dm3", "test.dm3.tile", copyout=copyout)

    '''
    
    def build(self, workfile, outfile, copyout=None):
        '''Main build function.'''
        # print "Building: %s -> %s"%(workfile, outfile)
        self.workfile = workfile
        self.outfile = outfile

        # Temporary directory
        self.tmpdir = tempfile.mkdtemp(prefix='emen2thumbs.')

        # Number of images in the file
        self.nimg = EMAN2.EMUtil.get_image_count(workfile)

        # Awful hack to write out regular thumbs
        self.copyout = copyout or {}

        # Build each image in the file
        ret = []        
        for index in range(self.nimg):
            ret.append(self._build(index))
        
        # Combine all the files into the tile file
        self._build_compile(ret, outfile)
        
        # CLeanup
        try:
            # print "Removing tmpdir:", self.tmpdir
            os.rmdir(self.tmpdir)
        except:
            # print "Couldn't remove tmpdir: ", self.tmpdir
            pass
        
        # print ret
        return ret
        

    def _build(self, index):
        '''Build a single image in the file.'''
        # print "...index:", index
        header = {}

        # Read the header
        img = EMAN2.EMData()
        img.read_image(self.workfile, 0, True)
        h = img.get_attr_dict()
        
        # Copy basic header information
        header['nx'] = h['nx']
        header['ny'] = h['ny']
        header['nz'] = h['nz']
        header['slices'] = []
        
        if header['nz'] == 1:
            # 2D Image
            img2 = EMAN2.EMData()
            img2.read_image(self.workfile, index, False)
            img2.process_inplace("normalize")
            if self.nimg > 1:
                # ... stack of 2D images.
                header['slices'].append(self.build_slice(img2, index=index, fixed=[128,512,1024]))
            else:
                # regular old 2D image -- also generate power spectrum and tiles.
                header['slices'].append(self.build_slice(img2, index=index, tile=True, pspec=True, fixed=[128,512,1024]))
        
        else:        
            # 3D Image -- read region for each Z slice
            for i in range(header['nz']):
                region = EMAN2.Region(0, 0, i, header['nx'], header['ny'], 1)
                img2 = EMAN2.EMData()
                img2.read_image(self.workfile, 0, False, region)
                header['slices'].append(self.build_slice(img2, index=index, nz=i, fixed=[128,512,1024]))
        
        return header
        

    def _build_compile(self, ret, outfile):
        '''Combine the results into a single tile file.'''
        # print "...compiling"    
        # Get items to pack
        self.pos = 0
        self.packed = []
        def pack(items):
            for i,j in sorted(items.items()):
                f = j[0]
                j[0] = self.pos
                j[1] = os.stat(f).st_size
                self.pos += j[1]
                self.packed.append(f)
        
        for index in ret:
            for sl in index.get('slices', []):
                pack(sl.get('tiles', {}))
                pack(sl.get('fixed', {}))
                pack(sl.get('pspec', {}))
                pack(sl.get('pspec1d', {}))
        
        # Move all the packed files into the archive
        tf = file(outfile, "wb")
        pickle.dump(ret, tf)
        for filename in self.packed:
            a = open(filename, "r")
            b = a.read()
            a.close()
            tf.write(b)
            os.remove(filename)

        tf.close()        


    def build_slice(self, img, nz=1, index=0, tile=False, pspec=False, fixed=None):
        '''Build a single 2D slice from a 2D or 3D image.'''
        header = {}

        h = img.get_attr_dict()
        header['nx'] = h['nx']
        header['ny'] = h['ny']
        header['nz'] = h['nz']
        
        if tile:
            # print "...tiles"
            header['tiles'] = self.build_tiles(img, nz=nz, index=index)

        if pspec:
            # print "...pspec"
            header['pspec'], header['pspec1d'] = self.build_pspec(img, nz=nz, index=index)

        if fixed:
            # print "...fixed"
            header['fixed'] = {}
            for f in fixed:
                header['fixed'][f] = self.build_fixed(img, tilesize=f, nz=nz, index=index)

        return header
        
        
    def build_tiles(self, img, nz=1, index=0, tilesize=256):
        '''Build tiles for a 2D slice.'''
        # Work with a copy of the EMData
        img2 = img.copy()        

        # Calculate the number of zoom levels based on the tile size
        levels = math.ceil( math.log( max(img.get_xsize(), img.get_ysize()) / tilesize) / math.log(2.0) )

        # Tile header
        header = img.get_attr_dict()
        tile_dict = {}

        # Step through shrink range creating tiles
        for level in range(int(levels)):
            scale = 2**level
            rmin = img2.get_attr("mean") - img2.get_attr("sigma") * 3.0
            rmax = img2.get_attr("mean") + img2.get_attr("sigma") * 3.0
            for x in range(0, img2.get_xsize(), tilesize):
                for y in range(0, img2.get_ysize(), tilesize):
                    i = img2.get_clip(EMAN2.Region(x, y, tilesize, tilesize), fill=rmax)
                    i.set_attr("render_min", rmin)
                    i.set_attr("render_max", rmax)
                    i.set_attr("jpeg_quality", 80)
                    # Write output
                    fsp = "tile.index-%d.scale-%d.z-%d.x-%d.y-%d.jpg"%(index, scale, nz, x/tilesize, y/tilesize)
                    fsp = os.path.join(self.tmpdir, fsp)
                    i.write_image(fsp)
                    tile_dict[(scale, x/tilesize, y/tilesize)] = [fsp, None, 'jpg', tilesize, tilesize]
                    
            # Shrink by 2 for next round.
            img2.process_inplace("math.meanshrink",{"n":2})

        return tile_dict

        
    def build_fixed(self, img, tilesize=256, nz=1, index=0):
        """Build a thumbnail of a 2D EMData."""
        # Output files
        fsp = "fixed.index-%d.z-%d.size-%d.jpg"%(index, nz, tilesize)
        fsp = os.path.join(self.tmpdir, fsp)

        # The scale factor
        thumb_scale = img.get_xsize() / float(tilesize), img.get_ysize() / float(tilesize)
        sc = 1 / max(thumb_scale)

        if tilesize == 0 or sc >= 1.0:
            # Write out a full size jpg
            img2 = img.copy()
        else:
            # Shrink the image
            # print "shrink to thumbnail with scale factor:", sc, 1/sc, math.ceil(1/sc)
            # img2 = img.process("xform.scale", {"scale":sc, "clip":tilesize})
            img2 = img.process("math.meanshrink", {'n':math.ceil(1/sc)})

        # Adjust the brightness for rendering
        rmin = img2.get_attr("mean") - img2.get_attr("sigma") * 3.0
        rmax = img2.get_attr("mean") + img2.get_attr("sigma") * 3.0
        img2.set_attr("render_min", rmin)
        img2.set_attr("render_max", rmax)
        img2.set_attr("jpeg_quality", 80)        
        img2.write_image(fsp)
        
        # Awful hack to write out regular thumbs
        if index == 0 and nz == 1 and tilesize in self.copyout:
            # print "...copy thumb:", tilesize, self.copyout[tilesize]
            img2.write_image(self.copyout[tilesize])

        return [fsp, None, 'jpg', img2.get_xsize(), img2.get_ysize()]

    
    def build_pspec(self, img, tilesize=512, nz=1, index=0):
        """Build a 2D FFT and 1D rotationally averaged power spectrum of a 2D EMData."""
        
        # Return dictionaries
        pspec_dict = {}
        pspec1d_dict = {}

        # Output files
        outfile = "pspec.index-%d.z-%d.size-%d.png"%(index, nz, tilesize)
        outfile = os.path.join(self.tmpdir, outfile)        

        outfile1d = "pspec1d.index-%d.z-%d.size-%d.json"%(index, nz, tilesize)
        outfile1d = os.path.join(self.tmpdir, outfile1d)        

        # Create a new image to hold the 2D FFT
        nx, ny = img.get_xsize() / tilesize, img.get_ysize() / tilesize
        a = EMAN2.EMData()
        a.set_size(tilesize, tilesize)
        
        # Image isn't big enough..
        if (ny<2 or nx<2):
            return pspec_dict, pspec1d_dict
            
        # Create FFT
        for y in range(1,ny-1):
            for x in range(1,nx-1):
                c = img.get_clip(EMAN2.Region(x*tilesize, y*tilesize, tilesize, tilesize))
                c.process_inplace("normalize")
                c.process_inplace("math.realtofft")
                c.process_inplace("math.squared")
                a += c

        # Reset the center value
        a.set_value_at(tilesize/2, tilesize/2, 0, .01)

        # Adjust brightness
        a -= a.get_attr("minimum") - a.get_attr("sigma") * .01
        a.process_inplace("math.log")
        a.set_attr("render_min", a.get_attr("minimum") - a.get_attr("sigma") * .1)
        a.set_attr("render_max", a.get_attr("mean") + a.get_attr("sigma") * 4.0)

        # Write out the PSpec png
        a.write_image(outfile)

        # Add to dictionary
        pspec_dict[tilesize] = [outfile, None, 'png', a.get_xsize(), a.get_ysize()]

        # Calculate
        t = (tilesize/2)-1
        y = a.calc_radial_dist(t, 1, 1, 0) # radial power spectrum (log)
        f = open(outfile1d, "w")
        json.dump(y,f)
        f.close()
        pspec1d_dict[tilesize] = [outfile1d, None, 'json', t]
        
        # Return both the 2D and 1D files
        return pspec_dict, pspec1d_dict


##### Cryo-EM file format handlers #####

class EMDataHandler(BinaryHandler):
    _allow_gzip = True

    def _thumbnail_build(self, workfile):
        outfile = self._outfile('eman2')
        copyout = {
            128: self._outfile('thumb.jpg'),
            512: self._outfile('small.jpg')
        }
        builder = EMDataBuilder()
        tile = builder.build(str(workfile), outfile, copyout=copyout)            
        


@BinaryHandler.register(['dm3', 'mrc', 'tif', 'tiff', 'hdf'])
class MicrographHandler(EMDataHandler):
    _allow_gzip = True
    


# IMPORTANT -- Do not change this.
if __name__ == "__main__":
    emen2.db.handlers.main(globals())

    



# $Id: query.py,v 1.26 2013/04/09 09:08:18 irees Exp $
import os
import operator
import datetime
import copy
import urllib

import emen2.util.listops
from emen2.web.view import View

cmp_order = [
    ".is.",
    "==",
    ".not.",
    "!=",
    ".contains.",
    ".gte.",
    ">=",
    ".lte.",
    "<=",
    ".gt.",
    ">",
    ".lt.",
    "<",
    ".any.",
    '.none.',
    '.noop.',
    '.name.',
]


def query_to_path(q):
    pass



def path_to_query(path, **kwargs):
    q = {}
    q['c'] = []
    if path == None:
        path = ''
    c = path.split("/")

    for constraint in c:
        constraint = urllib.unquote(constraint)
        match = []
        foundcomps = filter(lambda x:constraint.partition(x)[1], cmp_order)
        if foundcomps:
            comp = foundcomps[0]
            p1, p2, p3 = constraint.partition(comp)
            p1 = urllib.unquote(p1)
            p2 = urllib.unquote(p2).replace('.','')
            p3 = urllib.unquote(p3)

            if p1:
                match.append(p1)
                if p2:
                    match.append(p2)
                    if p3:
                        match.append(p3)

        elif constraint:
            match = [constraint]

        if match:
            q['c'].append(match)

            
    # General query...
    keywords = kwargs.pop('keywords', None)
    if keywords:
        q['c'].append(['*','contains',keywords])

    q.update(kwargs)
    return q



class TooManyFiles(Exception):
    pass


@View.register
class Query(View):

    def initq(self, path=None, q=None, c=None, **kwargs):
        self.template = '/pages/query.main'
        self.title = "Query"
        if q:
            self.q = q
        else:
            self.q = path_to_query(path, **kwargs)
        if c:
            self.q['c'] = c
        self.ctxt['parent'] = None
        self.ctxt['rectype'] = None
        self.ctxt['header'] = True
        self.ctxt['controls'] = True
        
    @View.add_matcher(r'^/query/$', name='main')
    @View.add_matcher(r'^/query/(?P<path>.*)/$', name='query')
    def main(self, path=None, q=None, c=None, **kwargs):
        self.initq(path, q, c, **kwargs)
        self.q = self.db.table(**self.q)
        self.ctxt['q'] = self.q

    @View.add_matcher(r'^/query/(?P<path>.*)/embed/$')
    def embed(self, path=None, q=None, c=None, create=False, rectype=None, parent=None, controls=True, header=True, **kwargs):
        # create/rectype/parent for convenience.
        self.main(path, q, c)
        self.template = '/pages/query'
        # awful hack
        self.ctxt['controls'] = controls
        self.ctxt['header'] = header
        self.ctxt['parent'] = parent
        self.ctxt['rectype'] = rectype
        
    @View.add_matcher(r'^/query/(?P<path>.*)/edit/$', name='edit')
    def edit(self, path=None, q=None, c=None, **kwargs):
        self.initq(path, q, c)
        self.template = '/pages/query.edit'
        self.q = self.db.table(**self.q)
        self.ctxt['q'] = self.q

    @View.add_matcher(r'^/plot/$', name='plot')
    @View.add_matcher(r'^/plot/(?P<path>.*)/$', name='plot_path')
    def plot(self, path=None, q=None, c=None, x=None, y=None, z=None, **kwargs):
        self.initq(path, q, c)
        self.q['x'] = x
        self.q['y'] = y
        self.q['z'] = z
        self.template = '/pages/query.plot'
        self.q = self.db.plot(**self.q)
        # print "Plot results:", self.q
        self.ctxt['q'] = self.q

    # /download/ can't be in the path because of a emen2resource.getchild issue
    @View.add_matcher(r'^/query/(?P<path>.*)/attachments/$', name='attachments')
    def attachments(self, path=None, q=None, c=None, confirm=False, **kwargs):
        self.initq(path, q, c)
        self.template = '/pages/query.attachments'
        self.q = self.db.query(**self.q)
        self.ctxt['q'] = self.q

        # Look up all the binaries
        bdos = self.db.binary.find(record=self.q['names'], count=0)
        if len(bdos) > 100000 and not confirm:
            raise TooManyFiles, "More than 100,000 files returned. Please see the admin if you need to download the complete set."

        records = set([i.record for i in bdos])
        users = set([bdo.get('creator') for bdo in bdos])
        users = self.db.user.get(users)
        self.ctxt['users'] = users
        self.ctxt['recnames'] = self.db.view(records)
        self.ctxt['bdos'] = bdos

    # /download/ can't be in the path because of a emen2resource.getchild issue
    @View.add_matcher(r'^/query/(?P<path>.*)/attachments2/$', name='attachments2')
    def attachments2(self, path=None, q=None, c=None, confirm=False, **kwargs):
        self.initq(path, q, c)
        self.template = '/pages/query.main'
        self.q = self.db.query(**self.q)        

        # Look up all the binaries
        bdos = self.db.binary.find(record=self.q['names'], count=0)
        names = [i.name for i in bdos]
        bdoq = self.db.table(subset=names, keytype="binary", checkbox=True, view="{{thumbnail(name)}} {{filename}} {{filesize}} {{recname(record)}}")

        # Render the binary table instead of the record table
        self.ctxt['q'] = bdoq





__version__ = "$Revision: 1.26 $".split(":")[1][:-1].strip()



# $Id: binary.py,v 1.59 2013/05/01 03:16:07 irees Exp $
"""Database support for Binary attachments."""

import time
import re
import traceback
import math
import os

# For file writing
import shutil
import hashlib
import cStringIO
import tempfile

# EMEN2 imports
import emen2.db.dataobject
import emen2.db.config
import emen2.db.exceptions


class Binary(emen2.db.dataobject.BaseDBObject):
    """Binary file stored on disk and managed by EMEN2.

    Provides following attributes:
        filename, record, md5, filesize, compress, filepath

    The Binary name has a specific format, bdo:YYYYMMDDXXXXX, where YYYYMMDD is
    date format and XXXXX is 5-char hex ID code of file for the day.

    The filename attribute is the original name of the uploaded file. The
    filesize attribute is the uncompressed size of the file, and the md5
    attribute is the MD5 checksum of the uncompressed file. The filename may be
    changed by an owner after a Binary is committed, but the contents (filesize
    and md5) cannot be changed. Files may be compressed after they have been
    created; in this case, the compress param will specify the compression
    scheme (gzip, bz2, etc.) and filesize_compress and md5_compress will have
    values for the compressed file.

    File names are checked to prevent illegal characters, and names such as "."
    and invalid file names on some platforms ("COM", "NUL", etc.).

    Binaries are generally associated with a Record, stored in the record
    attribute. Read permission on a Binary requires either ownership of the
    item, or read permission on the associated Record. The owner of the Binary
    can change which Record is associated.

    When a Binary is retrieved from the database, the filepath property will
    be accessible. This points to the physical file on disk containing the data.

    These BaseDBObject methods are extended:

        init            Set attributes
        setContext        Check read permissions and bind Context
        validate        Check required attributes

    And the following method is provided:

        parse            Parse name

    :attr filename: File name
    :attr filesize: Size of the uncompressed file
    :attr md5: MD5 checksum of the uncompressed file
    :attr filesize_compress: Size of the compressed file
    :attr md5_compress: MD5 checksum of the compressed file
    :attr compress: File is gzip compressed
    :attr record: Associated Record
    :property filepath: Path to the file on disk
    """

    attr_public = emen2.db.dataobject.BaseDBObject.attr_public | set(["filepath", "filename", "record", "compress", "filesize", "md5", "filesize_compress", "md5_compress"])
    attr_protected = emen2.db.dataobject.BaseDBObject.attr_protected | set(["compress", "filesize", "md5", "filesize_compress", "md5_compress"])
    filepath = property(lambda x:x._filepath)

    def init(self, d):
        self.__dict__['filename'] = None
        self.__dict__['record'] = None
        self.__dict__['md5'] = None
        self.__dict__['filesize'] = None
        self.__dict__['compress'] = False
        self.__dict__['filesize_compress'] = None
        self.__dict__['md5_compress'] = None
        self.__dict__['_filepath'] = None
        self._filepath = None

    ##### DBObject interface #####

    def setContext(self, ctx):
        super(Binary, self).setContext(ctx=ctx)
        self.__dict__['_filepath'] = self.parse(self.name).get('filepath')
        if self.isowner():
            return True
        if self.record is not None:
            rec = self._ctx.db.record.get(self.record, filt=False)

    # filepath is set during setContext, and discarded during commit (todo)
    def _set_filepath(self, key, value):
        return set()

    # These immutable attributes only ever be set for a new Binary, before commit
    def _set_md5(self, key, value):
        if self.name:
            raise emen2.db.exceptions.ValidationError, "Cannot change a Binary's file attachment"
        return self._set(key, value, self.isowner())

    def _set_md5_compress(self, key, value):
        if self.name:
            raise emen2.db.exceptions.ValidationError, "Cannot change a Binary's file attachment"
        return self._set(key, value, self.isowner())

    def _set_compress(self, key, value):
        if self.name:
            raise emen2.db.exceptions.ValidationError, "Cannot change a Binary's file attachment"
        return self._set(key, value, self.isowner())

    def _set_filesize(self, key, value):
        if self.name:
            raise emen2.db.exceptions.ValidationError, "Cannot change a Binary's file attachment"
        return self._set(key, value, self.isowner())

    def _set_filesize_compress(self, key, value):
        if self.name:
            raise emen2.db.exceptions.ValidationError, "Cannot change a Binary's file attachment"
        return self._set(key, value, self.isowner())

    # These can be changed normally
    def _set_filename(self, key, value):
        # Sanitize filename.. This will allow unicode characters,
        #    and check for reserved filenames on linux/windows
        value = unicode(value)
        value = "".join([i for i in value if i.isalpha() or i.isdigit() or i in '.()-=_'])
        if value.upper() in ['..', '.', 'CON', 'PRN', 'AUX', 'NUL',
                                    'COM1', 'COM2', 'COM3', 'COM4', 'COM5',
                                    'COM6', 'COM7', 'COM8', 'COM9', 'LPT1',
                                    'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6',
                                    'LPT7', 'LPT8', 'LPT9']:
            value = u"renamed."+value
        return self._set(key, value, self.isowner())

    def _set_record(self, key, value):
        return self._set(key, value, self.isowner())

    def validate(self):
        # Validate
        if not all([self.filename, self.md5, self.filesize != None]):
            raise emen2.db.exceptions.ValidationError, "Filename, filesize, and MD5 checksum are required"

    ##### Utility methods #####

    # Write contents to a temporary file.
    def writetmp(self, filedata=None, fileobj=None, basepath=None, suffix="upload"):
        '''Write to temporary storage, and calculate size/md5.
        :return: Temporary file path, the file size, and an md5 digest.
        '''
        if filedata:
            fileobj = cStringIO.StringIO(filedata)
        if not fileobj:
            raise ValueError, "No data to write to temporary file."

        fileobj.seek(0)

        # In narrow circumstances, this might not be the same filesystem
        # as the final file destination. So use a higher level tool like
        # shutil to rename the file, instead of just os.rename.
        basepath = basepath or self.parse('')['basepath']

        if not os.path.exists(basepath):
            os.makedirs(basepath)
            
        (fd, tmpfile) = tempfile.mkstemp(dir=basepath, suffix='.%s'%suffix)

        m = hashlib.md5()
        filesize = 0
        # Copy to the output file, updating md5 and size
        with os.fdopen(fd, "w+b") as f:
            for line in fileobj:
                f.write(line)
                m.update(line)
                filesize += len(line)

        md5sum = m.hexdigest()
        emen2.db.log.info("Wrote file: %s, filesize: %s, md5sum: %s"%(tmpfile, filesize, md5sum))

        # Update filesize/md5
        self.__dict__['filesize'] = filesize
        self.__dict__['md5'] = md5sum
        return tmpfile

    @staticmethod
    def parse(bdokey, counter=None):
        """Parse a 'bdo:2010010100001' type identifier into constituent parts
        and resolve location in the filesystem."""

        prot, _, bdokey = (bdokey or "").rpartition(":")
        if not prot:
            prot = "bdo"

        # ian: todo: implement other BDO protocols, e.g. references to uris
        if prot not in ["bdo"]:
            raise Exception, "Invalid binary protocol: %s"%prot

        if bdokey:
            # Now process; must be 14 chars long..
            year = int(bdokey[:4])
            mon = int(bdokey[4:6])
            day = int(bdokey[6:8])
            if counter == None:
                counter = int(bdokey[9:13],16)
        else:
            # Timestamps are now in ISO8601 format
            # e.g.: "2011-10-16T02:00:00+00:00"
            bdokey = emen2.db.database.utcnow()
            year = int(bdokey[:4])
            mon = int(bdokey[5:7])
            day = int(bdokey[8:10])
            counter = counter or 0

        # YYYYMMDD
        datekey = "%04d%02d%02d"%(year, mon, day)
        datedir = os.path.join('%04d'%year, '%02d'%mon, '%02d'%day)

        # Get the binary storage paths from config
        # Find the last item matching the current date
        binarypaths = emen2.db.config.get('paths.binary')
        bp = [x for x in sorted(binarypaths.keys()) if str(x)<=datekey]

        # Resolve the parsed bdo key to a directory (basepath) and file (filepath)
        basepath = os.path.join(binarypaths[bp[-1]], datedir)
        filepath = os.path.join(basepath, "%05X"%counter)
        
        # ... same for previewpath
        previewpaths = emen2.db.config.get('paths.preview')
        pp = [x for x in sorted(previewpaths.keys()) if str(x)<=datekey]
        previewpath = os.path.join(previewpaths[pp[-1]], datedir, "%05X"%counter)

        # The BDO name (bdo:YYYYMMDDXXXXX)
        name = "%s:%s%05X"%(prot, datekey, counter)

        return {
            "prot":prot,
            "year":year,
            "mon":mon,
            "day":day,
            "counter":counter,
            "datekey":datekey,
            "basepath":basepath,
            "filepath":filepath,
            "previewpath":previewpath,
            "name":name
            }


__version__ = "$Revision: 1.59 $".split(":")[1][:-1].strip()


# $Id: btrees.py,v 1.158 2013/05/14 15:33:23 irees Exp $
"""BerkeleyDB Driver."""

import collections
import copy
import functools
import os
import shutil
import sys
import time
import traceback
import weakref
import cPickle as pickle

# Berkeley DB
# Note: the 'bsddb' module is not sufficient.
import bsddb3

# EMEN2 imports
import emen2.db.config
import emen2.db.log
import emen2.db.query
import emen2.util.listops
from emen2.db.exceptions import *

# EMEN2 DBObjects
import emen2.db.dataobject
import emen2.db.record
import emen2.db.binary
import emen2.db.paramdef
import emen2.db.recorddef
import emen2.db.user
import emen2.db.context
import emen2.db.group

Record = emen2.db.record.Record
Binary = emen2.db.binary.Binary
ParamDef = emen2.db.paramdef.ParamDef
RecordDef = emen2.db.recorddef.RecordDef
User = emen2.db.user.User
NewUser = emen2.db.user.NewUser
Group = emen2.db.group.Group
Context = emen2.db.context.Context

try:
    import emen2.db.bulk
    bulk = emen2.db.bulk
    # emen2.db.log.info("Note: using EMEN2-BerkeleyDB bulk access module")
except ImportError, inst:
    bulk = None


# ian: todo: move this to EMEN2DBEnv
DB_CONFIG = """\
# Don't touch these
set_data_dir data
set_lg_dir journal
set_lg_regionmax 1048576
set_lg_max 8388608
set_lg_bsize 2097152
"""


##### EMEN2 Database Environment #####

class EMEN2DBEnv(object):
    """BerkeleyDB Database Environment."""

    def __init__(self, path=None, snapshot=False):
        """
        :keyword path: Directory containing environment.
        :keyword snapshot: Use Berkeley DB Snapshot (Multiversion Concurrency Control) for read transactions
        """
        
        # Database environment directory
        self.path = path
        self.snapshot = snapshot or (not emen2.db.config.get('params.snapshot'))
        self.cachesize = emen2.db.config.get('bdb.cachesize') * 1024 * 1024l

        # Make sure the data directory exists.
        paths = [
            os.path.join(self.path, 'data'),
            os.path.join(self.path, 'journal')
        ]
        for p in paths:
            if not os.path.exists(p):
                os.makedirs(p)

        # Make sure the DB_CONFIG is present.
        configpath = os.path.join(self.path, "DB_CONFIG")
        exists = os.path.exists(configpath)
        if not exists:
            emen2.db.log.debug("BDB: Copying default DB_CONFIG file: %s"%configpath)
            f = open(configpath, "w")
            f.write(DB_CONFIG)
            f.close()

        # Databases
        self.keytypes =  {}

        # Pre- and post-commit actions.
        # These are used for things like renaming files during the commit phase.
        # TODO: The details of this are highly likely to change,
        #     or be moved to a different place.
        self._txncbs = collections.defaultdict(list)

        # Open DBEnv and main tables.
        self.dbenv = self.open()
        self.init()
        
    def _load_json(self, infile, keytypes=None, ctx=None, txn=None):
        """Load a JSON file containing DBOs."""
        # Create a special root context to load the items
        loader = emen2.db.load.BaseLoader(infile=infile)
        keytypes = keytypes or ['paramdef', 'user', 'group', 'recorddef', 'binary', 'record']
        for keytype in keytypes:
            for item in loader.loadfile(keytype=keytype):
                emen2.db.log.debug("CREATE: %s %s"%(keytype, item.get('name')))
                print item
                i = self[keytype].dataclass(ctx=ctx)
                i._load(item)
                self[keytype]._put_data(i.name, i, txn=txn)

    def create(self):
        emen2.db.log.info("BDB: Loading from JSON")

        # Start txn
        ctx = emen2.db.context.SpecialRootContext()
        txn = self.newtxn(write=True)
        keytypes = ['paramdef', 'recorddef']
        
        # Load core ParamDefs... Items defined in base.json are required.
        infile = emen2.db.config.get_filename('emen2', 'db/base.json')
        self._load_json(infile, keytypes=keytypes, ctx=ctx, txn=txn)

        # Load DBOs from extensions.
        emen2.db.config.load_jsons(cb=self._load_json, keytypes=keytypes, ctx=ctx, txn=txn)
        
        # Rebuild indexes
        self['paramdef']._rebuild_indexes(ctx=ctx, txn=txn)
        self['recorddef']._rebuild_indexes(ctx=ctx, txn=txn)

        # Commit txn
        self.txncommit(txn=txn)

    
    def open(self):
        """Open the Database Environment."""
        emen2.db.log.info("BDB: Opening database environment: %s"%self.path)
        dbenv = bsddb3.db.DBEnv()

        if self.snapshot:
            dbenv.set_flags(bsddb3.db.DB_MULTIVERSION, 1)
        
        txncount = (self.cachesize / 4096) * 2
        if txncount > 1024*128:
            txncount = 1024*128
            
        dbenv.set_cachesize(0, self.cachesize)
        dbenv.set_tx_max(txncount)
        dbenv.set_lk_max_locks(300000)
        dbenv.set_lk_max_lockers(300000)
        dbenv.set_lk_max_objects(300000)

        flags = 0
        flags |= bsddb3.db.DB_CREATE
        flags |= bsddb3.db.DB_INIT_MPOOL
        flags |= bsddb3.db.DB_INIT_TXN
        flags |= bsddb3.db.DB_INIT_LOCK
        flags |= bsddb3.db.DB_INIT_LOG
        flags |= bsddb3.db.DB_THREAD
        dbenv.open(self.path, flags)
        return dbenv

    def init(self):
        # Authentication. These are not public.
        self._context = CollectionDB(dataclass=Context, dbenv=self)
        # These are public dbs.
        self.keytypes['paramdef']  = CollectionDB(dataclass=ParamDef, dbenv=self)
        self.keytypes['recorddef'] = CollectionDB(dataclass=RecordDef, dbenv=self)
        self.keytypes['group']     = CollectionDB(dataclass=Group, dbenv=self)
        self.keytypes['record']    = RecordDB(dataclass=Record, dbenv=self)
        self.keytypes['user']      = UserDB(dataclass=User, dbenv=self)
        self.keytypes['newuser']   = NewUserDB(dataclass=NewUser, dbenv=self)
        self.keytypes['binary']    = BinaryDB(dataclass=Binary, dbenv=self)

    # ian: todo: make this nicer.
    def close(self):
        """Close the Database Environment"""
        for k,v in self.keytypes.items():
            v.close()
        self.dbenv.close()

    def __getitem__(self, key, default=None):
        """Pass dictionary gets to self.keytypes."""
        return self.keytypes.get(key, default)

    ##### Transaction management #####

    def newtxn(self, write=False):
        """Start a new transaction.
        
        :keyword write: Transaction will be likely to write data; turns off Berkeley DB Snapshot
        :return: New transaction
        """
        flags = bsddb3.db.DB_TXN_SNAPSHOT
        if write:
            flags = 0

        txn = self.dbenv.txn_begin(flags=flags)
        emen2.db.log.debug("TXN: start: %s flags %s"%(txn, flags))
        return txn

    def txncheck(self, txn=None, write=False):
        """Check a transaction status, or create a new transaction.

        :keyword txn: An existing open transaction
        :keyword write: See newtxn
        :return: Open transaction
        """
        if not txn:
            txn = self.newtxn(write=write)
        return txn

    def txnabort(self, txn):
        """Abort transaction.

        :keyword txn: An existing open transaction
        :exception: KeyError if transaction was not found
        """
        emen2.db.log.debug("TXN: abort: %s"%txn)
        txnid = txn.id()
        self._txncb(txnid, 'abort', 'before')
        txn.abort()
        self._txncb(txnid, 'abort', 'after')
        self._txncbs.pop(txnid, None)

    def txncommit(self, txn):
        """Commit a transaction.

        :param txn: An existing open transaction
        :exception: KeyError if transaction was not found
        """
        emen2.db.log.debug("TXN: commit: %s"%txn)
        txnid = txn.id()
        self._txncb(txnid, 'commit', 'before')
        txn.commit()
        self._txncb(txnid, 'commit', 'after')
        self._txncbs.pop(txnid, None)

    ##### Pre- and post-commit hooks. #####

    def txncb(self, txn, action, args=None, kwargs=None, condition='commit', when='before'):
        # Add a pre- or post-commit hook.
        if when not in ['before', 'after']:
            raise ValueError, "Transaction callback 'when' must be before or after"
        if condition not in ['commit', 'abort']:
            raise ValueError, "Transaction callback 'condition' must be commit or abort"
        item = [condition, when, action, args or [], kwargs or {}]
        self._txncbs[txn.id()].append(item)

    # This is still being developed. Do not touch.
    def _txncb(self, txnid, condition, when):
        # Note: this takes txnid, not a txn. This
        # is because txn.id() is null after commit.
        actions = self._txncbs.get(txnid, [])
        for c, w, action, args, kwargs in actions:
            if w == when and c == condition:
                if action == 'rename':
                    self._txncb_rename(*args, **kwargs)
                elif action == 'email':
                    self._txncb_email(*args, **kwargs)
                elif action == 'thumbnail':
                    self._txncb_thumbnail(*args, **kwargs)
    
    def _txncb_rename(self, source, dest):
        emen2.db.log.info("Renaming file: %s -> %s"%(source, dest))
        try:
            shutil.move(source, dest)
        except Exception, e:
            emen2.db.log.error("Couldn't rename file %s -> %s"%(source, dest))

    def _txncb_email(self, *args, **kwargs):
        try:
            sendmail(*args, **kwargs)
        except Exception, e:
            emen2.db.log.error("Couldn't send email: %s"%e)
            
    def _txncb_thumbnail(self, bdo):
        try:
            emen2.db.handlers.thumbnail_from_binary(bdo, wait=False)
        except Exception, e:
            emen2.db.log.error("Couldn't start thumbnail builder")
    
    ##### Log archive #####

    def checkpoint(self, txn=None):
        """Checkpoint the database environment."""
        return self.dbenv.txn_checkpoint()

    def journal_archive(self, remove=True, checkpoint=True, txn=None):
        """Archive completed log files.

        :keyword remove: Remove the log files after moving them to the backup location
        :keyword checkpoint: Run a checkpoint first; this will allow more files to be archived
        """
        outpath = emen2.db.config.get('paths.journal_archive')

        if checkpoint:
            emen2.db.log.info("BDB: Log Archive: Checkpoint")
            self.dbenv.txn_checkpoint()

        archivefiles = self.dbenv.journal_archive(bsddb3.db.DB_ARCH_ABS)

        emen2.db.log.info("BDB: Log Archive: Preparing to move %s completed log files to %s"%(len(archivefiles), outpath))

        if not os.access(outpath, os.F_OK):
            os.makedirs(outpath)

        outpaths = []
        for archivefile in archivefiles:
            dest = os.path.join(outpath, os.path.basename(archivefile))
            emen2.db.log.info('BDB: Log Archive: %s -> %s'%(archivefile, dest))
            shutil.move(archivefile, dest)
            outpaths.append(dest)

        return outpaths     


# Berkeley DB wrapper classes
class BaseDB(object):
    """BerkeleyDB B-tree based DB.
    
    Note this class does not define access methods. It just create, opens, and closes the DB.
    
    :attr filename: Filename of BDB on disk
    :attr dbenv: EMEN2 Database Environment
    :attr bdb: Berkeley DB instance
    :attr DBOPENFLAGS: Berkeley DB flags for opening database
    :attr DBSETFLAGS: Additional flags
    """

    def __init__(self, filename, keyformat='str', dataformat='str', dataclass=None, dbenv=None, extension='bdb'):
        """Create and open the DB.
        
        :param filename: Base filename to use
        :keyword dbenv: Database environment
        """
        # Filename
        self.filename = filename
        self.extension = extension
        
        # EMEN2DBEnv
        self.dbenv = dbenv

        # What are we storing?
        # This sets formatters and types for keys and data.
        self._setkeyformat(keyformat)
        self._setdataformat(dataformat, dataclass)

        # BDB handle and open flags.
        self.bdb = None
        self.DBOPENFLAGS = bsddb3.db.DB_AUTO_COMMIT | bsddb3.db.DB_THREAD | bsddb3.db.DB_CREATE
        self.DBSETFLAGS = []

        # Init and open.
        self.init()
        self.open()

    def init(self):
        """Subclass init hook."""
        pass

    ##### DB methods #####

    def open(self):
        """Open the DB. This uses an implicit open transaction."""
        if self.bdb:
            raise Exception, "DB already open"

        # Create the DB handle and set flags
        self.bdb = bsddb3.db.DB(self.dbenv.dbenv)

        # Set DB flags, e.g. duplicate keys allowed
        for flag in self.DBSETFLAGS:
            self.bdb.set_flags(flag)

        # Open the DB with the correct flags.
        emen2.db.log.debug("BDB: %s open"%self.filename)
        fn = '%s.%s'%(self.filename, self.extension)
        self.bdb.open(filename=fn, dbtype=bsddb3.db.DB_BTREE, flags=self.DBOPENFLAGS)

    def close(self):
        """Close the DB."""
        emen2.db.log.debug("BDB: %s close"%self.filename)
        self.bdb.close()
        self.bdb = None

    # Dangerous!
    def truncate(self, txn=None, flags=0):
        """Truncate BDB (e.g. 'drop table'). Transaction required.
        :keyword txn: Transaction
        """
        # todo: Do more checking before performing a dangerous operation.
        emen2.db.log.debug("BDB: %s truncate"%self.filename)
        self.bdb.truncate(txn=txn)
    
    ##### load/dump methods for keys and data #####

    # DO NOT TOUCH THIS!!!!
    def _cfunc_numeric(self, k1, k2):
        # Numeric comparison function, for key sorting.
        if not k1:
            k1 = 0
        else:
            k1 = self.keyload(k1)

        if not k2:
            k2 = 0
        else:
            k2 = self.keyload(k2)
        return cmp(k1, k2)

    # def _get_json(self, key, txn=None, flags=0):
    #     pass
    #     
    # def _get_data(self, key, txn=None, flags=0):
    #     pass
    #     
    # def _put_json(self, key, value, txn=None, flags=0):
    #     pass
    # 
    # def _put_pickle(self, key, value, txn=None, flags=0):
    #     pass

    def _setkeyformat(self, keyformat):
        # Set the DB key type. This will bind the correct
        # keyclass, keydump, keyload methods.
        if keyformat == 'str':
            self.keyclass = unicode
            self.keydump = lambda x:unicode(x).encode('utf-8')
            self.keyload = lambda x:x.decode('utf-8')
        elif keyformat == 'int':
            self.keyclass = int
            self.keydump = str
            self.keyload = int
        elif keyformat == 'float':
            self.keyclass = float
            self.keydump = lambda x:pickle.dumps(x)
            self.keyload = lambda x:pickle.loads(x or 'N.')
        else:
            raise ValueError, "Invalid key format: %s. Supported: str, int, float"%keyformat
        self.keyformat = keyformat

    def _setdataformat(self, dataformat, dataclass=None):
        # Set the DB data type. This will bind the correct
        # dataclass attribute, and datadump and dataload methods.
        if dataclass:
            dataformat = 'pickle'
        if dataformat == 'str':
            # String dataformat; use UTF-8 encoded strings.
            self.dataclass = unicode
            self.datadump = lambda x:unicode(x).encode('utf-8')
            self.dataload = lambda x:x.decode('utf-8')
        elif dataformat == 'int':
            # Decimal dataformat, use str encoded ints.
            self.dataclass = int
            self.datadump = str
            self.dataload = int
        elif dataformat == 'float':
            # Float dataformat; these do not sort natively, so pickle them.
            self.dataclass = float
            self.datadump = lambda x:pickle.dumps(x)
            self.dataload = lambda x:pickle.loads(x or 'N.')
        elif dataformat == 'pickle':
            # This DB stores a DBO as a pickle.
            self.dataclass = dataclass
            self.datadump = lambda x:pickle.dumps(x)
            self.dataload = lambda x:pickle.loads(x or 'N.')
        else:
            # Unknown dataformat.
            raise Exception, "Invalid data format: %s. Supported: str, int, float, pickle"%dataformat
        self.dataformat = dataformat
    

class IndexDB(BaseDB):
    '''EMEN2DB optimized for indexes.

    Security is NOT checked here, so do not expose indexes directly to 
    untrusted clients. A user could query values they could not otherwise
    access.

    IndexDB uses the Berkeley DB facility for storing multiple values for a
    single key (DB_DUPSORT). The Berkeley DB API has a method for
    quickly reading these multiple values.

    This class is intended for use with an OPTIONAL C module, _bulk.so, that
    accelerates reading from the index. The Berkeley DB bulk reading mode
    is not fully implemented in the bsddb3 package; the C module does the bulk
    reading in a single function call, greatly speeding up performance, and
    returns the correct native Python type. The C module is totally optional
    and is transparent; the only change is read speed.

    In the DBEnv directory, IndexDBs will have a ".index" extension.

    Index references are added using addrefs() and removerefs(). These both
    take a single key, and a list of references to add or remove.

    Extends or overrides the following methods:
        init        Checks bulk mode
        get         Returns all values found.
        keys        Index keys
        items       Index items; (key, [value1, value2, ...])
        iteritems   Index iteritems

    Adds the following indexing methods:
        addrefs        Add (key, [values]) references to the index
        removerefs    Remove (key, [values]) references from the index

    '''

    def init(self):
        """Open DB with support for duplicate keys."""
        self.DBSETFLAGS = [bsddb3.db.DB_DUPSORT]
        self._setbulkmode(True)
        super(IndexDB, self).init()

    def _setbulkmode(self, bulkmode):
        # Use acceleration C module if available
        self._get_method = self._get_method_nonbulk
        if bulk:
            if bulkmode:
                self._get_method = emen2.db.bulk.get_dup_bulk
            else:
                self._get_method = emen2.db.bulk.get_dup_notbulk

    def _get_method_nonbulk(self, cursor, key, dt, flags=0):
        # Get without C module. Uses an already open cursor.
        n = cursor.set(key)
        r = set() #[]
        m = cursor.next_dup
        while n:
            r.add(n[1])
            n = m()
        return set(self.dataload(x) for x in r)

    # Default get method used by get()
    _get_method = _get_method_nonbulk

    def get(self, key, default=None, cursor=None, txn=None, flags=0):
        """Return all the values for this key.

        Can be passed an already open cursor, or open one if necessary.
        Requires a transaction. The real get method is _get_method, which
        is set during init based on availability of the C module.

        :param key: Key
        :keyword default: Default value if key not found
        :keyword cursor: Use this cursor
        :keyword txn: Transaction
        :return: Values for key

        """
        emen2.db.log.debug("BDB: %s get: %s"%(self.filename, key))        
        if cursor:
            r = self._get_method(cursor, self.keydump(key), self.dataformat)
        else:
            cursor = self.bdb.cursor(txn=txn)
            r = self._get_method(cursor, self.keydump(key), self.dataformat)
            cursor.close()
        if bulk and self.dataformat == 'pickle':
            r = set(self.dataload(x) for x in r)
        return r

    # ian: todo: allow min/max
    def keys(self, minkey=None, maxkey=None, txn=None, flags=0):
        """Keys. Transaction required.

        :keyword txn: Transaction

        """
        keys = set(map(self.keyload, self.bdb.keys(txn)))
        return list(keys)

    # ian: todo: allow min/max
    def items(self, minkey=None, maxkey=None, txn=None, flags=0):
        """Accelerated items. Transaction required.

        :keyword txn: Transaction

        """
        ret = []
        cursor = self.bdb.cursor(txn=txn)
        pair = cursor.first()
        while pair != None:
            data = self._get_method(cursor, pair[0], self.dataformat)
            if bulk and self.dataformat == "pickle":
                data = set(map(self.dataload, data))
            ret.append((self.keyload(pair[0]), data))
            pair = cursor.next_nodup()
        cursor.close()
        return ret

    def iteritems(self, minkey=None, maxkey=None, txn=None, flags=0):
        """Iteritems. Transaction required.

        :keyword minkey: Minimum key
        :keyword maxkey: Maximum key
        :keyword txn: Transaction
        :yield: (key, value)

        """
        ret = []
        cursor = self.bdb.cursor(txn=txn)
        pair = cursor.first()

        # Start a minimum key.
        # This only works well if the keys are sorted properly.
        if minkey is not None:
            pair = cursor.set_range(self.keydump(minkey))
        while pair != None:
            data = self._get_method(cursor, pair[0], self.dataformat)
            k = self.keyload(pair[0])
            if bulk and self.dataformat == "pickle":
                data = set(map(self.dataload, data))
            yield (k, data)
            pair = cursor.next_nodup()
            if maxkey is not None and k > maxkey:
                pair = None
        cursor.close()

    ##### Write Methods #####

    def removerefs(self, key, items, txn=None):
        '''Remove references.

        :param key: Key
        :param items: References to remove
        :keyword txn: Transaction
        :return: Keys that no longer have any references

        '''
        if not items: return []
        emen2.db.log.debug("BDB: %s removerefs: %s -> %s"%(self.filename, key, items))
        delindexitems = []

        cursor = self.bdb.cursor(txn=txn)

        key = self.keyclass(key)
        items = map(self.dataclass, items)

        dkey = self.keydump(key)
        ditems = map(self.datadump, items)

        for ditem in ditems:
            if cursor.set_both(dkey, ditem):
                cursor.delete()

        if not cursor.set(dkey):
            delindexitems.append(key)

        cursor.close()
        return delindexitems

    def addrefs(self, key, items, txn=None):
        """Add references.

        A list of keys that are new to this index are returned. This can be
        used to maintain other indexes.

        :param key: Key
        :param items: References to add
        :keyword txn: Transaction
        :return: Keys that are new to this index

        """
        if not items: return []
        emen2.db.log.debug("BDB: %s addrefs: %s -> %s"%(self.filename, key, items))
        addindexitems = []

        key = self.keyclass(key)
        items = map(self.dataclass, items)
        
        dkey = self.keydump(key)
        ditems = map(self.datadump, items)
        
        cursor = self.bdb.cursor(txn=txn)

        if not cursor.set(dkey):
            addindexitems.append(key)

        for ditem in ditems:
            try:
                cursor.put(dkey, ditem, flags=bsddb3.db.DB_KEYFIRST)
            except bsddb3.db.DBKeyExistError, e:
                pass

        cursor.close()

        return addindexitems



# Context-aware DB for Database Objects.
# These support a single DB and a single data class.
# Supports sequenced items.
class CollectionDB(BaseDB):
    '''Database for items supporting the DBO interface (mapping
    interface, setContext, writable, etc. See BaseDBObject.)

    You may consider terms collection, bucket, keytype, and table to be
    interchangeable.

    Most methods require a transaction. Additionally, because
    this class manages DBOs, most methods also require a Context.

    Extends the following methods:
        __init__         Opens DBs in a specific directory
        open             Also opens sequencedb
        close            Also cloes sequencdb and indexes

    And adds the following methods that require a context:
        new              New item
        get              Get a single item
        gets             Get items
        put              Put a single item
        puts             Put items
        filter           Context-aware keys
        query            Query
        validate         Validate an item
        exists           Check if an item exists already

    May be deprecated, since they're unbounded:
        keys
        values
        items

    Some internal methods:
        _get_data
        _put
        _puts
        _put_data
        _exists

    Sequence methods:
        _update_names    Update items with new names from sequence
        _key_generator
        _incr_sequence
        _get_max

    Index methods:
        getindex         Open an index
        _reindex         Calculate index updates
        _reindex_*       Write index updates
    
    Relationship methods:
        parents          Returns parents
        children         Returns children
        siblings         Item siblings
        rel              General relationship method
        pclink           Add a parent/child relationship
        pcunlink         Remove a parent/child relationship
        relink           Add and remove several relationships
    '''
    
    def __init__(self, *args, **kwargs):
        # Change the filename slightly
        dataclass = kwargs.get('dataclass')
        dbenv = kwargs.get('dbenv')
        self.keytype = (kwargs.get('keytype') or dataclass.__name__).lower()
        
        # Sequences
        self.sequencedb = None
        
        # Indexes
        self.indexes = {}
        self._truncate_index = False

        filename = os.path.join(self.keytype, self.keytype)
        d1 = os.path.join(dbenv.path, 'data', self.keytype)
        d2 = os.path.join(dbenv.path, 'data', self.keytype, 'index')
        for i in [d1, d2]:
            try: os.makedirs(i)
            except: pass  
        return super(CollectionDB, self).__init__(filename, *args, **kwargs)

    def open(self):
        """Open DB, and sequence."""
        super(CollectionDB, self).open()
        self.sequencedb = bsddb3.db.DB(self.dbenv.dbenv)
        self.sequencedb.open(os.path.join('%s.sequence.bdb'%self.filename), dbtype=bsddb3.db.DB_BTREE, flags=self.DBOPENFLAGS)

    def close(self):
        """Close DB, sequence, and indexes."""
        super(CollectionDB, self).close()
        self.sequencedb.close()
        self.sequencedb = None
        for k,v in self.indexes.items():
            if v:
                ind.close()
        self.indexes = {}

    ##### New items.. #####

    def new(self, *args, **kwargs):
        """Returns new DBO. Requires ctx and txn.

        All the method args and keywords will be passed to the constructor.

        :keyword txn: Transaction
        :return: New DBO
        :exception ExistingKeyError:
        
        """
        txn = kwargs.pop('txn', None) # Don't pass the txn..
        ctx = kwargs.get('ctx', None)
        kwargs['keytype'] = self.keytype
        inherit = kwargs.pop('inherit', [])
        item = self.dataclass(*args, **kwargs)

        for i in inherit:
            # Allow to raise an exception if does not exist or cannot read.
            i = self.get(i, filt=False, ctx=ctx, txn=txn)
            if i.get('permissions'):
                item.addumask(i.get('permissions'))
            if i.get('groups'):
                item.addgroup(i.get('groups'))
            item['parents'].add(i.name)

        # Acquire a write lock on this name.
        if self.exists(item.name, txn=txn, flags=bsddb3.db.DB_RMW):
            raise emen2.db.exceptions.ExistingKeyError, "%s already exists"%item.name

        return item

    ##### Exists #####
    
    def exists(self, key, ctx=None, txn=None, flags=0):
        """Check if a key exists."""
        # Names that are None or a negative int will be automatically assigned.
        # In this case, return immediately and don't acquire any locks.
        # Note: this method does not check permissions; you could use it to check
        #     if a key exists or not, even if you can't read the value.
        emen2.db.log.debug("BDB: %s exists: %s"%(self.filename, key))        
        if key < 0 or key is None:
            return False
        return self.bdb.exists(self.keydump(key), txn=txn, flags=flags)
        
    ##### Keys, values, items #####
    
    def filter(self, names=None, ctx=None, txn=None):
        """Filter a set of keys for read permission.

        :keyword names: Subset of items to check
        :keyword ctx: Context
        :keyword txn: Transaction
        :return: Set of keys that are accessible by the Context

        """
        if names is not None:
            if ctx.checkreadadmin():
                return names
            items = self.gets(names, ctx=ctx, txn=txn)
            return set([i.name for i in items])
        return set(self.keys(txn=txn))
    
    def keys(self, ctx=None, txn=None):
        emen2.db.log.warn("BDB: %s keys: Deprecated method!"%self.filename)
        return map(self.keyload, self.bdb.keys(txn))
    
    def items(self, ctx=None, txn=None):
        emen2.db.log.warn("BDB: %s items: Deprecated method!"%self.filename)
        ret = []
        for k,v in self.bdb.items(txn):
            i = self.dataload(v)
            i.setContext(ctx)
            ret.append((self.keyload(k), i))
        return ret
        
    def values(self, ctx=None, txn=None):
        raise NotImplementedError

    ##### Filtered context gets.. #####

    def get(self, key, filt=True, ctx=None, txn=None, flags=0):
        """See cgets(). This works the same, but for a single key."""
        r = self.gets([key], txn=txn, ctx=ctx, filt=filt, flags=flags)
        if not r:
            return None
        return r[0]

    def gets(self, keys, filt=True, ctx=None, txn=None, flags=0):
        """Get a list of items, with a Context. Requires ctx and txn.

        The filt keyword, if True, will ignore KeyError and SecurityError.
        Alternatively, it can be set to a list of Exception types to ignore.

        :param key: Items to get
        :keyword filt: Ignore KeyError, SecurityError
        :keyword ctx: Context
        :keyword txn: Transaction
        :return: DBOs with bound Context
        :exception KeyError:
        :exception SecurityError:

        """
        if filt == True:
            filt = (emen2.db.exceptions.SecurityError, KeyError)

        ret = []
        for key in keys:
            try:
                d = self._get_data(key, txn=txn, flags=flags)
                d.setContext(ctx)
                ret.append(d)
            except filt, e:
                pass
        return ret
        
    def _get_data(self, key, txn=None, flags=0):
        emen2.db.log.debug("BDB: %s get: %s"%(self.filename, key))        
        kd = self.keydump(key)
        d = self.dataload(self.bdb.get(kd, txn=txn, flags=flags))
        if d:
            return d
        raise KeyError, "No such key %s"%(key)    

    def _get_data_items(self, txn=None, flags=0):
        # items() without ctx.
        for k,v in self.bdb.items(txn=txn, flags=flags):
            k = self.keyload(key)
            v = self.dataload(v)
            yield k,v

    ##### Write methods #####

    def validate(self, items, ctx=None, txn=None):
        return self.puts(items, commit=False, ctx=ctx, txn=txn)

    def put(self, item, commit=True, ctx=None, txn=None):
        """See puts(). This works the same, but for a single DBO."""
        ret = self.puts([item], commit=commit, ctx=ctx, txn=txn)
        if not ret:
            return None
        return ret[0]
        
    def puts(self, items, commit=True, ctx=None, txn=None):
        """Update DBOs. Requires ctx and txn.

        :param item: DBOs, or similar (e.g. dict)
        :keyword commit: Actually commit (e.g. for validation only)
        :keyword ctx: Context
        :keyword txn: Transaction
        :return: Updated DBOs
        :exception KeyError:
        :exception SecurityError:
        :exception ValidationError:

        """
        # Updated items
        crecs = []
        for updrec in items:
            name = updrec.get('name')
            
            # Get the existing item or create a new one.
            if self.exists(name, txn=txn, flags=bsddb3.db.DB_RMW):
                # Get the existing item.
                orec = self._get_data(name, txn=txn, flags=bsddb3.db.DB_RMW)
                # May raise a SecurityError if you can't read it.
                orec.setContext(ctx)
                orec.update(updrec)
            else:
                # Create a new item.
                orec = self.new(ctx=ctx, txn=txn, **updrec)

            # Update the item.
            orec.validate()
            crecs.append(orec)

        # If we just wanted to validate the changes, return before writing changes.
        if commit:
            return self._puts(crecs, ctx=ctx, txn=txn)
        return crecs
        
    def _put(self, item, ctx=None, txn=None):
        return self._puts([item], ctx=ctx, txn=txn)[0]

    def _puts(self, items, ctx=None, txn=None):
        # TODO: ctx used only for cache.
        # Assign names for new items.
        # This will also update any relationships to uncommitted records.
        self._update_names(items, txn=txn)

        # Now that names are assigned, calculate the index updates.
        ind = self._reindex(items, txn=txn)

        # Write the items "for real."
        for item in items:
            self._put_data(item.name, item, txn=txn)

        # Write index updates
        self._reindex_write(ind, ctx=ctx, txn=txn)
        emen2.db.log.debug("BDB: Committed %s items"%(len(items)))
        return items

    def _put_data(self, name, item, txn=None, flags=0):
        emen2.db.log.debug("BDB: %s put: %s"%(self.filename, item.name))        
        self.bdb.put(self.keydump(name), self.datadump(item), txn=txn, flags=flags)
    
    # def delete(self, name, ctx=None, txn=None, flags=0):
    #     emen2.db.log.debug("BDB: %s put: %s"%(self.filename, item.name))        
    #     self.bdb.delete(self.keydump(name), txn=txn, flags=flags)


    def query(self, c=None, mode='AND', subset=None, ctx=None, txn=None):
        """Return a Query Constraint Group.

        You will need to call constraint.run() to execute the query,
        and constraint.sort() to sort the values.
        """
        return emen2.db.query.Query(constraints=c, mode=mode, subset=subset, ctx=ctx, txn=txn, btree=self)

    ##### Changes to indexes #####
    
    def _reindex(self, items, reindex=False, txn=None):
        """Update indexes.

        The original items will be retrieved and compared to the updated
        items. A set of index changes will be calculated, and then handed off
        to the _reindex_* methods. In some cases, these will be overridden
        or extended to handle special indexes -- such as parent/child
        relationships in RelateDB.
        
        :param items: Updated DBOs
        :keyword reindex:
        :keyword ctx: Context
        :keyword txn: Transaction

        """
        # Updated indexes
        ind = collections.defaultdict(list)

        # Get changes as param:([name, newvalue, oldvalue], ...)
        for item in items:
            # Get the current record for comparison of updated values.
            # Use an empty dict for new records so all keys
            # will seen as new (or if reindexing)
            if item.isnew() or reindex:
                orec = {}
            else:
                orec = self._get_data(item.name, txn=txn) or {}

            for param in item.changedparams(orec):
                ind[param].append((item.name, item.get(param), orec.get(param)))

        # Return the index changes.
        return ind

    # .... the actual items need to be written ^^^ between these two vvv steps for relationships to be updated ....

    def _reindex_write(self, ind, ctx=None, txn=None):
        """(Internal) Write index updates."""
        # Parent/child relationships are a special case.
        # The other side of the relationship needs to be updated. 
        # Calculate the correct changes here, but do not
        # update the indexes yet. 
        # Update the parent child relationships.
        
        parents = ind.pop('parents', None)
        children = ind.pop('children', None)
        self._reindex_relink(parents, children, txn=txn)

        # Now, Update indexes.
        for k,v in ind.items():
            self._reindex_param(k, v, ctx=ctx, txn=txn)

    def _reindex_param(self, param, changes, ctx=None, txn=None):
        """(Internal) Reindex changes to a single parameter."""
        # TODO: ctx used only for cache...
        # If nothing changed, skip.
        if not changes:
            return

        # If we can't access the index, skip. (Raise Exception?)
        ind = self.getindex(param, txn=txn)
        if ind == None:
            return

        # Check that this key is currently marked as indexed
        # Check the cache for the param
        hit, pd = ctx.cache.check(('paramdef', param))
        if not hit:
            pd = self.dbenv['paramdef']._get_data(param, txn=txn)
            ctx.cache.store(('paramdef', param), pd)
        vt = emen2.db.vartypes.Vartype.get_vartype(pd.vartype, pd=pd, db=ctx.db, cache=ctx.cache)

        # Process the changes into index addrefs / removerefs
        try:
            addrefs, removerefs = vt.reindex(changes)
        except Exception, e:
            # print "Could not reindex param %s: %s"%(pd.name, e)
            # print changes
            return

        # Write!
        for oldval, recs in removerefs.items():
            ind.removerefs(oldval, recs, txn=txn)
        for newval,recs in addrefs.items():
            ind.addrefs(newval, recs, txn=txn)

    ##### Manage indexes. #####

    def getindex(self, param, txn=None):
        """Open a parameter index. Requires txn.

        A successfully opened IndexDB will be cached in self.indexes[param] and
        reused on subsequent calls.

        If an index for a parameter isn't returned by this method, the reindex
        method will just skip it.

        :param param: Parameter
        :param txn: Transaction
        :return: IndexDB

        """
        if param in self.indexes:
            return self.indexes.get(param)

        # Check the paramdef to see if it's indexed.
        try:
            pd = self.dbenv["paramdef"]._get_data(param, txn=txn)
        except KeyError:
            self.indexes[param] = None
            return None
        
        # Check the key format.
        vartype = emen2.db.vartypes.Vartype.get_vartype(pd.vartype, pd=pd)
        tp = vartype.keyformat
        if not pd.indexed or not tp:
            self.indexes[param] = None
            return None
            
        # Open the index
        indname = os.path.join(self.keytype, 'index', param)
        ind = emen2.db.btrees.IndexDB(filename=indname, keyformat=tp, dataformat=self.keyformat, dbenv=self.dbenv)

        # Cache the open index.
        self.indexes[param] = ind
        
        # Careful; truncate all indexes if this is set. Used for rebuilding all indexes.
        if self._truncate_index and ind:
           ind.truncate(txn=txn) 

        return ind

    def _rebuild_indexes(self, ctx=None, txn=None):
        emen2.db.log.info("BDB: Rebuilding indexes: Start")
        # ugly hack..
        self._truncate_index = True
        for k in self.indexes:
            self.indexes[k].truncate(txn=txn)

        # Do this in chunks of 1,000 items
        # Get all the keys -- do not include cached items
        keys = sorted(map(self.keyload, self.bdb.keys(txn)), reverse=True)
        for chunk in emen2.util.listops.chunk(keys, 1000):
            if chunk:
                emen2.db.log.info("BDB: Rebuilding indexes: %s ... %s"%(chunk[0], chunk[-1]))
            items = [self._get_data(i, txn=txn) for i in chunk]
            # Use self.reindex() instead of self.puts() -- the data should
            # already be validated, so we can skip that step.
            # self.puts(items, txn=txn)
            ind = self._reindex(items, reindex=True, txn=txn)
            self._reindex_write(ind, ctx=ctx, txn=txn)

        self._truncate_index = False
        emen2.db.log.info("BDB: Rebuilding indexes: Done")

    ##### Sequences #####

    # Todo: Simplify this. Maybe move it somewhere else.
    def _update_names(self, items, txn=None):
        """Update items with new names. Requires txn.

        :param items: Items to update.
        :keyword txn: Transaction.

        """
        namemap = {}
        for item in items:
            if not self.exists(item.name, txn=txn):
                # Get a new name.
                newname = self._key_generator(item, txn=txn)
                try:
                    newname = self.keyclass(newname)
                except:
                    raise Exception, "Invalid name: %s"%newname
                # Check the name is still available, and acquire lock.
                if self.exists(newname, txn=txn, flags=bsddb3.db.DB_RMW):
                    raise emen2.db.exceptions.ExistingKeyError, "%s already exists"%newname
                # Update the item's name.
                namemap[item.name] = newname
                item.__dict__['name'] = newname

        # Update all the record's links
        for item in items:
            item.__dict__['parents'] = set([namemap.get(i,i) for i in item.parents])
            item.__dict__['children'] = set([namemap.get(i,i) for i in item.children])

        return namemap

    def _key_generator(self, item, txn=None):
        # Set name policy in this method.
        return unicode(item.name or emen2.db.database.getrandomid())

    def _incr_sequence(self, key='sequence', txn=None):
        # Update a sequence key. Requires txn.
        # The Sequence DB can handle multiple keys -- e.g., for
        # binaries, each day has its own sequence key.
        delta = 1
        
        val = self.sequencedb.get(key, txn=txn, flags=bsddb3.db.DB_RMW)
        if val == None:
            val = 0
        val = int(val)
        
        self.sequencedb.put(key, str(val+delta), txn=txn)
        emen2.db.log.debug("BDB: %s sequence: %s -> %s"%(self.filename, val, val+delta))
        return val

    def _get_max(self, key="sequence", txn=None):
        """Return the current maximum item in the sequence. Requires txn.

        :keyword txn: Transaction
        """
        sequence = self.sequencedb.get(key, txn=txn)
        if sequence == None:
            sequence = 0
        val = int(sequence)
        return val

    ##### Relationship methods #####

    def parents(self, names, recurse=1, ctx=None, txn=None):
        """See rel(), with rel='parents", tree=False. Requires ctx and txn.

        This will return a dict of parents to the specified recursion depth.

        :return: Dict with names as keys, and their parents as values

        """
        return self.rel(names, recurse=recurse, rel='parents', ctx=ctx, txn=txn)

    def children(self, names, recurse=1, ctx=None, txn=None):
        """See rel(), with rel="children", tree=False. Requires ctx and txn.

        This will return a dict of children to the specified recursion depth.

        :return: Dict with names as keys, and their children as values

        """
        return self.rel(names, recurse=recurse, rel='children', ctx=ctx, txn=txn)

    # Siblings
    def siblings(self, name, ctx=None, txn=None):
        """Siblings. Note this only takes a single name. Requries ctx and txn.

        :keyword name: DBO name
        :keyword ctx: Context
        :keyword txn: Transaction
        :return: Set of siblings

        """
        parents = self.rel([name], rel='parents', ctx=ctx, txn=txn)
        allparents = set()
        for k,v in parents.items():
            allparents |= v
        siblings = set()
        children = self.rel(allparents, ctx=ctx, txn=txn)
        for k,v in children.items():
            siblings |= v
        return siblings

    # Checks permissions, return formats, etc..
    def rel(self, names, recurse=1, rel='children', tree=False, ctx=None, txn=None):
        """Find relationships. Requires context and transaction.

        Find relationships to a specified recusion depth. This supports any
        type of relationship that has a correctly setup index available;
        currently parents and children are supported. In the future, it will
        support any IndexDB that has names for keys, and the relationships
        as values.

        This method is public because it is sometimes convenient to find
        relationships based on a supplied argument without a case switch
        or duplication of code. However, it switches return types based on
        the tree keyword. Because of this complexity, it is usually called
        through the following convenience methods:
            parents, children, tree

        If tree keyword is True, the returned value will be a tree structure.
        This will have each specified DBO name as a key, and one level of
        children as values. These children will in turn have their own keys,
        and their own children as values, up to the specified recursion depth.

        If tree keyword is False, the returned value will be a dictionary
        with DBO names as keys, and their children (up to the specified
        recursion depth) as values.

        Example edges:
            1: 2, 3
            2: 3
            3: 4
            4: None

        ...with names = [1,2,3] and tree=True and recurse=-1:
            {1: [2,3], 2: [3], 3:[4], 4:[]}

        ...with names = [1,2,3], tree=False, and recurse=-1:
            {1: [2,3,4], 2: [3,4], 3:[4]}

        :keyword names: DBO names
        :keyword recurse: Recursion depth (default is 1)
        :keyword rel: Relationship type (default is children)
        :keyword tree: Set return type to tree or set
        :keyword ctx: Context
        :keyword txn: Transaction
        :return: Return a tree structure if tree=True, otherwise a set

        """
        result = {}
        visited = {}
        t = time.time()
        for i in names:
            result[i], visited[i] = self._bfs(i, rel=rel, recurse=recurse)

        # Flatten the dictionary to get all touched names
        allr = set()
        for v in visited.values():
            allr |= v

        # Filter by permissions
        allr = self.filter(allr, ctx=ctx, txn=txn)

        # If Tree=True, we're returning the tree... Filter for permissions.
        if tree:
            outret = {}
            for k, v in result.iteritems():
                for k2 in v:
                    outret[k2] = result[k][k2] & allr
            return outret

        # Else we're just ruturning the total list of all children,
        # keyed by requested record name
        for k in visited:
            visited[k] &= allr

        return visited

    def pclink(self, parent, child, ctx=None, txn=None):
        """Create parent-child relationship. Requires ctx and txn.

        Both items will have their parent/child attributes updated
        to reflect the new relationship. You must specify a Context, and
        have READ permissions on BOTH items, and WRITE permission on AT LEAST
        ONE item.

        :param parent: Parent
        :param child: Child
        :keyword ctx: Context
        :keyword txn: Transaction

        """
        self._putrel(parent, child, mode='addrefs', ctx=ctx, txn=txn)

    def pcunlink(self, parent, child, ctx=None, txn=None):
        """Remove parent-child relationship. Requires ctx and txn.

        Both items will have their parent/child attributes updated
        to reflect the deleted relationship. You must specify a Context, and
        have READ permissions on BOTH items, and WRITE permission on AT LEAST
        ONE item.

        :param parent: Parent
        :param child: Child
        :keyword ctx: Context
        :keyword txn: Transaction

        """
        self._putrel(parent, child, mode='removerefs', ctx=ctx, txn=txn)

    def relink(self, removerels=None, addrels=None, ctx=None, txn=None):
        """Add and remove a number of parent-child relationships at once."""
        removerels = removerels or []
        addrels = addrels or []
        remove = collections.defaultdict(set)
        add = collections.defaultdict(set)
        ci = emen2.util.listops.check_iterable

        for k,v in removerels.items():
            for v2 in ci(v):
                remove[self.keyclass(k)].add(self.keyclass(v2))
        for k,v in addrels.items():
            for v2 in ci(v):
                add[self.keyclass(k)].add(self.keyclass(v2))

        items = set(remove.keys()) | set(add.keys())
        items = self.gets(items, ctx=ctx, txn=txn)
        for item in items:
            item.children -= remove[item.name]
            item.children |= add[item.name]

        return self.puts(items, ctx=ctx, txn=txn)

    def _putrel(self, parent, child, mode='addrefs', txn=None):
        # (Internal) Add or remove a relationship.
        # Mode is addrefs or removerefs; it maps to the IndexDB method.

        # Check that we have enough permissions to write to one item
        # Use raw get; manually setContext. Allow KeyErrors to raise.
        p = self._get_data(parent, txn=txn, flags=bsddb3.db.DB_RMW)
        c = self._get_data(child, txn=txn, flags=bsddb3.db.DB_RMW)
        perm = []

        # Both items must exist, and we need to be able to write to one
        try:
            p.setContext(ctx)
            perm.append(p.writable())
        except emen2.db.exceptions.SecurityError:
            pass

        try:
            c.setContext(ctx)
            perm.append(c.writable())
        except emen2.db.exceptions.SecurityError:
            pass

        if not any(perm):
            raise emen2.db.exceptions.SecurityError, "Insufficient permissions to add/remove relationship"

        # Transform into the right format for _reindex_relink..
        newvalue = set() | p.children # copy
        if mode == 'addrefs':
            newvalue |= set([c.name])
        elif mode == 'removerefs':
            newvalue -= set([c.name])

        # The values will actually be set on the records
        #  during the relinking method..
        self._reindex_relink([], [[p.name, newvalue, p.children]], txn=txn)

    # Handle the reindexing...
    def _reindex_relink(self, parents, children, txn=None):
        # (Internal) Relink relationships
        # This method will grab both items, and add or remove the rels from
        # each item, and then update the parents/children IndexDBs.
        indc = self.getindex('children', txn=txn)
        indp = self.getindex('parents', txn=txn)

        # The names of new items.
        names = []

        # Process change sets into new and removed links
        add = []
        remove = []
        for name, new, old in (parents or []):
            old = set(old or [])
            new = set(new or [])
            for i in new - old:
                add.append((i, name))
            for i in old - new:
                remove.append((i, name))
            names.append(name)

        for name, new, old in (children or []):
            old = old or set()
            new = new or set()
            for i in new - old:
                add.append((name, i))
            for i in old - new:
                remove.append((name, i))
            names.append(name)

        p_add = collections.defaultdict(set)
        p_remove = collections.defaultdict(set)
        c_add = collections.defaultdict(set)
        c_remove = collections.defaultdict(set)

        for p,c in add:
            p_add[c].add(p)
            c_add[p].add(c)
        for p,c in remove:
            p_remove[c].add(p)
            c_remove[p].add(c)

        # Go and fetch other items that we need to update
        names = set(p_add.keys()+p_remove.keys()+c_add.keys()+c_remove.keys())
        # print "All affected items:", names
        # Get and modify the item directly w/o Context:
        # Linking only requires write permissions
        # on ONE of the items.
        for name in names:
            try:
                rec = self._get_data(name, txn=txn)
            except:
                continue
            rec.__dict__['parents'] -= p_remove[rec.name]
            rec.__dict__['parents'] |= p_add[rec.name]
            rec.__dict__['children'] -= c_remove[rec.name]
            rec.__dict__['children'] |= c_add[rec.name]
            self._put_data(rec.name, rec, txn=txn)
        for k,v in p_remove.items():
            if v:
                indp.removerefs(k, v, txn=txn)
        for k,v in p_add.items():
            if v:
                indp.addrefs(k, v, txn=txn)
        for k,v in c_remove.items():
            if v:
                indc.removerefs(k, v, txn=txn)
        for k,v in c_add.items():
            if v:
                indc.addrefs(k, v, txn=txn)
        return

    ##### Search tree-like indexes (e.g. parents/children) #####

    def _bfs(self, key, rel='children', recurse=1, ctx=None, txn=None):
        # (Internal) Tree search
        # Return a dict of results as well as the nodes visited (saves time)
        
        # Check max recursion depth
        if recurse == -1:
            recurse = emen2.db.config.get('params.maxrecurse')

        # Get the index, and create a cursor (slightly faster)
        rel = self.getindex(rel, txn=txn)
        cursor = rel.bdb.cursor(txn=txn)

        # Starting items

        # NOTE: I am using this ugly direct call because it saves 10-20% time.
        new = rel._get_method(cursor, rel.keydump(key), rel.dataformat) #

        stack = [new]
        result = {key: new}
        visited = set()
        lookups = []

        for x in xrange(recurse-1):
            if not stack[x]:
                break

            stack.append(set())
            for key in stack[x] - visited:
                new = rel._get_method(cursor, rel.keydump(key), rel.dataformat) 
                if new:
                    stack[x+1] |= new #.extend(new)
                    result[key] = new

            visited |= stack[x]

        visited |= stack[-1]
        cursor.close()
        return result, visited


class BinaryDB(CollectionDB):
    """CollectionDB for Binaries"""
    def _key_generator(self, item, txn=None):
        """Assign a name based on date, and the counter for that day."""
        # Get the current date and counter.
        dkey = emen2.db.binary.Binary.parse('')
        # Increment the day's counter.
        counter = self._incr_sequence(key=dkey['datekey'], txn=txn)
        # Make the new name.
        newdkey = emen2.db.binary.Binary.parse(dkey['name'], counter=counter)
        # Update the item's filepath..
        item.__dict__['_filepath'] = newdkey['filepath']
        # Return the new name.
        return newdkey['name']


ALLOW_RECORD_NAMES = False
class RecordDB(CollectionDB):
    def _key_generator(self, item, txn=None):
        # Set name policy in this method.
        if ALLOW_RECORD_NAMES:
            return unicode(item.name or emen2.db.database.getrandomid())
        return unicode(self._incr_sequence(txn=txn))

    # Todo: integrate with main filter method, since this works
    # for all permission-defined items.
    def filter(self, names, ctx=None, txn=None):
        """Filter for permissions.
        :param names: Record name(s).
        :returns: Readable Record names.
        """
        if names is None:
            if ctx.checkreadadmin():
                m = self._get_max(txn=txn)
                return set(map(unicode, range(0, m)))
                # return set(self.keys(txn=txn))
            ind = self.getindex("permissions", txn=txn)
            indc = self.getindex('creator', txn=txn)
            indg = self.getindex("groups", txn=txn)
            ret = ind.get(ctx.username, set(), txn=txn)
            ret |= indc.get(ctx.username, set(), txn=txn)
            for group in sorted(ctx.groups, reverse=True):
                ret |= indg.get(group, set(), txn=txn)
            return ret            

        names = set(names)

        if ctx.checkreadadmin():
            return names

        # If less than a thousand items, get directly.
        if len(names) <= 1000:
            items = self.gets(names, ctx=ctx, txn=txn)
            return set([i.name for i in items])

        # Make a copy
        find = copy.copy(names)

        # Use the permissions/groups index
        ind = self.getindex('permissions', txn=txn)
        indc = self.getindex('creator', txn=txn)
        indg = self.getindex('groups', txn=txn)

        find -= ind.get(ctx.username, set(), txn=txn)
        find -= indc.get(ctx.username, set(), txn=txn)
        for group in sorted(ctx.groups):
            if find:
                find -= indg.get(group, set(), txn=txn)

        return names - find


class UserDB(CollectionDB):
    def new(self, *args, **kwargs):
        txn = kwargs.get('txn', None)

        # DB.new. This will check the main bdb for an existing name.
        user = super(UserDB, self).new(*args, **kwargs)

        # Check  if this email already exists
        indemail = self.getindex('email', txn=txn)
        if indemail.get(user.email, txn=txn):
            raise emen2.db.exceptions.ExistingKeyError

        return user

    def filter(self, names=None, ctx=None, txn=None):
        # You need to be logged in to view this.
        if not ctx or ctx.username == 'anonymous':
            return set()
        return super(UserDB, self).filter(names, ctx=ctx, txn=txn)


class NewUserDB(CollectionDB):
    def delete(self, key, ctx=None, txn=None, flags=0):
        if not ctx.checkadmin():
            raise emen2.db.exceptions.SecurityError, "Only admin can delete keys."
        self.bdb.delete(self.keydump(key), txn=txn, flags=flags)

    def new(self, *args, **kwargs):
        txn = kwargs.get('txn', None)
        newuser = super(NewUserDB, self).new(*args, **kwargs)

        # Check if any pending accounts have this email address
        # for k,v in self._get_data_items(txn=txn):
        #    if newuser.email == v.email:
        #        raise emen2.db.exceptions.ExistingKeyError

        # Check  if this email already exists
        indemail = self.getindex('email', txn=txn)
        if indemail.get(newuser.email, txn=txn):
            raise emen2.db.exceptions.ExistingKeyError

        # Check if this email already exists
        indemail = self.dbenv["user"].getindex('email', txn=txn)
        if self.dbenv["user"].exists(newuser.name, txn=txn) or indemail.get(newuser.email, txn=txn):
            raise emen2.db.exceptions.ExistingKeyError

        return newuser

    def filter(self, names=None, ctx=None, txn=None):
        # This requires admin access
        if not ctx or not ctx.checkadmin():
            raise emen2.db.exceptions.SecurityError, "Admin rights needed to view user queue"
        return super(NewUserDB, self).filter(names, ctx=ctx, txn=txn)


__version__ = "$Revision: 1.158 $".split(":")[1][:-1].strip()


# $Id: vartypes.py,v 1.155 2013/05/14 15:33:23 irees Exp $
"""Vartypes (data types)

Classes:
    Vartype
    vt_*: A number of built-in data types

"""

import operator
import collections
import urllib
import re

# Working with time is even more fun.
import time
import calendar
import datetime

# ... dateutil helps.
import dateutil
import dateutil.parser
import dateutil.tz
import dateutil.rrule

# Markdown and Markupsafe are required now.
import markdown
from markupsafe import Markup, escape

# EMEN2 imports
import emen2.db.log
import emen2.db.exceptions
import emen2.util.listops

# Convenience
tzutc = dateutil.tz.tzutc()
ci = emen2.util.listops.check_iterable
ValidationError = emen2.db.exceptions.ValidationError

# Allow references to missing items.
ALLOW_MISSING = True

# Text values equivalent to "None"
NONEVALUES = [None, "", "N/A", "n/a", "None"]

##### Helper methods #####

def update_user_cache(cache, db, values):
    # Check cache
    to_cache = []
    for v in values:
        hit, dn = cache.check(('user', v))
        if not hit:
            to_cache.append(v)

    if to_cache:
        users = db.user.get(to_cache)
        for user in users:
            cache.store(('user', user.name), user)



def iso8601duration(d):
    """
    Parse ISO 8601 duration format.

    From Wikipedia, ISO 8601 duration format is:
        P[n]Y[n]M[n]DT[n]H[n]M[n]S

    P is the duration designator (historically called "period") placed at the start of the duration representation.
    Y is the year designator that follows the value for the number of years.
    M is the month designator that follows the value for the number of months.
    W is the week designator that follows the value for the number of weeks.
    D is the day designator that follows the value for the number of days.
    T is the time designator that precedes the time components of the representation.
    H is the hour designator that follows the value for the number of hours.
    M is the minute designator that follows the value for the number of minutes.
    S is the second designator that follows the value for the number of seconds.

    Examples:
    d = 'P1M2D' # 1 month, 2 days
    d = 'P1Y2M3DT4H5M6S' # 1 year, 2 months, 3 days, 4 hours, 5 minutes, 6 seconds
    d = 'P3W' # 3 weeks

    """

    regex = re.compile("""
            (?P<type>.)
            ((?P<weeks>[0-9,]+)W)?
            ((?P<years>[0-9,]+)Y)?
            ((?P<months>[0-9,]+)M)?
            ((?P<days>[0-9,]+)D)?
            (T
                ((?P<hours>[0-9,]+)H)?
                ((?P<minutes>[0-9,]+)M)?
                ((?P<seconds>[0-9,]+)S)?
            )?
            """, re.X)
    match = regex.search(d)
    rd = {} # return date

    # rdate['type'] = match.group('type')
    for key in ['weeks','years','months','days','hours','minutes','seconds']:
        if match.group(key):
            rd[key] = int(match.group(key))

    return rd


##### Vartypes #####

class Vartype(object):
    '''Base class for vartypes
    
    render
    validate
    reindex
    '''

    #: The index key type for this class
    keyformat = 'str'

    #: Is this vartype iterable?
    iterable = True
    
    #: Sort using rendered values?
    sort_render = False

    def __init__(self, pd=None, cache=None, db=None, options=None):
        self.pd = pd
        self.cache = cache
        self.db = db
        self.options = options or {}


    ##### Extensions #####

    registered = {}
    @classmethod
    def register(cls, name):
        def f(o):
            if name in cls.registered:
                raise ValueError("""%s is already registered""" % name)
            cls.registered[name] = o
            return o
        return f

    @classmethod
    def get_vartype(cls, name, *args, **kwargs):
        return cls.registered[name](*args, **kwargs)


    ##### Validation #####

    # def validate(self, pd, value):
    #     if value in NONEVALUES:
    #         return None
    # 
    #     if pd.property:
    #         value = self.properties[pd.property]().validate(self, pd, value, self.db)
    # 
    #     return self.vartypes[pd.vartype](cache=self.cache, db=self.db, pd=pd).validate(value)

    def validate(self, value):
        """Validate a value"""
        raise ValidationError, "%s is an organizational parameter, and is not intended to be used."%self.pd.name

    def _validate_reference(self, value, keytype=None):
        ret = []
        changed = False
        key = ('%s.names'%keytype,)
        hit, found = self.cache.check(key)
        if not hit:
            found = set()
            changed = True

        for i in value:       
            i = self.db._db.dbenv[keytype].keyclass(i) # ugly hack :(
            if i in found:
                ret.append(i)
            elif self.db.exists(i, keytype=keytype):
                ret.append(i)
                found.add(i)
                changed = True
            elif ALLOW_MISSING:
                emen2.db.log.warn("Validation: Could not find, but allowing: %s (parameter %s)"%(i, self.pd.name))
                ret.append(i)
            else:
                raise ValidationError, "Could not find: %s (parameter %s)"%(i, self.pd.name)
        
        if changed:
            self.cache.store(key, found)    
        return ret

    def _rci(self, value):
        """If the parameter is non-iterable, return a single value."""
        if value and not self.pd.iter:
            return value.pop()
        return value or None


    ##### Indexing #####

    def reindex(self, items):
        addrefs = collections.defaultdict(set)
        delrefs = collections.defaultdict(set)
        for name, new, old in items:
            if new == old:
                continue

            if not self.pd.iter:
                new=[new]
                old=[old]

            new = set(new or [])
            old = set(old or [])
            for n in new-old:
                addrefs[n].add(name)
            for o in old-new:
                delrefs[o].add(name)

        if None in addrefs:
            del addrefs[None]
        if None in delrefs:
            del delrefs[None]

        return addrefs, delrefs


    ##### Rendering #####

    def process(self, value):
        return value

    def render(self, value):
        output = self.options.get('output', 'unicode')
        r = self.render_unicode
        if output == "html":
            r = self.render_html
        elif output == "form":
            r = self.render_form
        return r(value)

    def render_unicode(self, value):
        """Render unicode values."""
        if value is None:
            return ''
        if self.pd.iter:
            return ", ".join(self._unicode(i) for i in value)
        return self._unicode(value)
    
    def render_html(self, value):
        """Render HTML formatted values. All values MUST be escaped."""
        if value is None:
            return ''
        if self.pd.iter:
            return self._li_wrap([self._html(i) for i in value])
        return self._html(value)

    def render_form(self, value):
        if self.pd.immutable:
            return self._html(value)
        if self.pd.iter:
            elem = self._li_wrap(
                [self._form(i) for i in (value or [])],
                hidden=self._form(None),
                add=self._add()
                )
        else:
            elem = self._form(value)
        return elem

    def _unicode(self, value):
        return unicode(value)
    
    def _html(self, value):
        if value is None:
            return ''
        elem = Markup("""<span class="e2-edit" data-paramdef="%s">%s</span>""")%(self.pd.name, self._unicode(value))
        return elem
        
    def _form(self, value):
        if value is None:
            value = ''
        elem = Markup("""<span class="e2-edit" data-paramdef="%s"><input type="text" name="%s" value="%s" /></span>""")%(
            self.pd.name,
            self.pd.name,
            value
        )
        return elem

    def _add(self):
        return """<input type="button" value="+" class="e2-edit-add" />"""

    def _li_wrap(self, values, hidden=None, add=None):
        lis = ["""<li>%s</li>"""%value for value in values]
        if hidden:
            lis.append("""<li class="e2-edit-template e2l-hide">%s</li>"""%hidden)            
        if add:
            lis.append("""<li>%s</li>"""%add)
        return """<ul>%s</ul>"""%"".join(lis)
        


@Vartype.register('none')
class vt_none(Vartype):
    pass
    
    
# Float vartypes
#    Indexed as 'f'
@Vartype.register('float')
class vt_float(Vartype):
    """Floating-point number."""

    keyformat = 'float'

    def validate(self, value):
        return self._rci(map(float, ci(value)))

    def _unicode(self, value):
        u = self.pd.defaultunits or ''
        return '%g %s'%(value, u)

    def _form(self, value):    
        if value is None:
            value = ''
        units = self.pd.defaultunits or ''
        return Markup("""<span class="e2-edit" data-paramdef="%s"><input type="text" name="%s" value="%s" /> %s</span>""")%(
            self.pd.name,
            self.pd.name,
            value,
            units
        )
    

@Vartype.register('percent')
class vt_percent(Vartype):
    """Percentage. 0 <= x <= 1."""

    keyformat = 'float'

    def validate(self, value):
        value = map(float, ci(value))
        for i in value:
            if not 0 <= i <= 1.0:
                raise ValidationError, "Range for percentage is 0 <= value <= 1.0; value was: %s"%i
        return self._rci(value)

    def _unicode(self, value):
        return '%0.0f'%(i*100.0)



# Integer vartypes
#    Indexed as 'd'
@Vartype.register('int')
class vt_int(Vartype):
    """Integer."""

    keyformat = 'int'

    def validate(self, value):
        return self._rci(map(int, ci(value)))


@Vartype.register('coordinate')
class vt_coordinate(Vartype):
    """Coordinates; tuples of floats."""

    keyformat = None

    def validate(self, value):
        return [(float(x), float(y)) for x,y in ci(value)]


@Vartype.register('boolean')
class vt_boolean(Vartype):
    """Boolean value. Accepts 0/1, True/False, T/F, Yes/No, Y/N, None."""

    keyformat = 'int'

    def validate(self, value):
        t = ['t', 'y', 'yes', 'true', '1']
        f = ['f', 'n', 'no', 'false', 'none', '0']
        ret = []
        for i in ci(value):
            i = unicode(value).lower()
            if i in t:
                i = True
            elif i in f:
                i = False
            else:
                raise ValidationError, "Invalid boolean: %s"%unicode(value)
            ret.append(i)
        return self._rci(ret)

    def _form(self, value):
        choices = []
        if value is None:
            choices.append("""<option value="" checked="checked" />""")
            choices.append("""<option>True</option>""")
            choices.append("""<option>False</option>""")
        elif value:
            choices.append("""<option value="" />""")
            choices.append("""<option checked="checked">True</option>""")
            choices.append("""<option>False</option>""")
        else:
            choices.append("""<option value="" />""")
            choices.append("""<option>True</option>""")
            choices.append("""<option checked="checked" >False</option>""")
        return Markup("""<span class="e2-edit" data-paramdef="%s"><select>%s</select></span>""")%(
            self.pd.name, 
            Markup("".join(choices))
            )
        


# String vartypes
#    Indexed as keyformat 'str'
@Vartype.register('string')
class vt_string(Vartype):
    """String."""

    def validate(self, value):
        return self._rci([unicode(x).strip() for x in ci(value)])


@Vartype.register('choice')
class vt_choice(vt_string):
    """One value from a defined list of choices."""

    def validate(self, value):
        value = [unicode(i).strip() for i in ci(value)]
        for v in value:
            if v not in self.pd.choices:
                raise ValidationError, "Invalid choice: %s"%v
        return self._rci(value)

    def _form(self, value):
        choices = []
        if value is None:
            value = ''
        for choice in ['']+self.pd.choices:
            if choice == value:
                elem = Markup("""<option value="%s" selected="selected">%s</option>""")%(choice, choice)
            else:
                elem = Markup("""<option value="%s">%s</option>""")%(choice, choice)
            choices.append(elem)
            
        return Markup("""<span class="e2-edit" data-paramdef="%s"><select name="%s">%s</select></span>""")%(
            self.pd.name,
            self.pd.name,
            Markup("".join(choices))
            )
        


@Vartype.register('recorddef')
class vt_recorddef(vt_string):
    """RecordDef name."""

    def validate(self, value):
        value = self._validate_reference(ci(value), keytype='recorddef')
        return self._rci(value)



@Vartype.register('text')
class vt_text(vt_string):
    """Freeform text, with word indexing."""

    elem = 'div'
    unindexed_words = set(["in", "of", "for", "this", "the", "at", "to", "from", "at", "for", "and", "it", "or"])

    def reindex(self, items):
        """(Internal) calculate param index updates for vartype == text"""
        addrefs = collections.defaultdict(list)
        delrefs = collections.defaultdict(list)
        for item in items:
            if item[1]==item[2]:
                continue

            for i in self._reindex_getindexwords(item[1]):
                addrefs[i].append(item[0])

            for i in self._reindex_getindexwords(item[2]):
                delrefs[i].append(item[0])

        allwords = set(addrefs.keys() + delrefs.keys()) - set(self.unindexed_words)
        addrefs2 = {}
        delrefs2 = {}

        for i in allwords:
            # make set, remove unchanged items
            addrefs2[i] = set(addrefs.get(i,[]))
            delrefs2[i] = set(delrefs.get(i,[]))
            u = addrefs2[i] & delrefs2[i]
            addrefs2[i] -= u
            delrefs2[i] -= u

        return addrefs2, delrefs2

    _reindex_getindexwords_m = re.compile('([a-zA-Z]+)|([0-9][.0-9]+)')
    def _reindex_getindexwords(self, value, ctx=None, txn=None):
        """(Internal) Split up a text param into components"""
        if value == None: return []
        value = unicode(value).lower()
        return set((x[0] or x[1]) for x in self._reindex_getindexwords_m.findall(value))

    def _form(self, value):
        if value is None:
            value = ''
        return Markup("""<div class="e2-edit" data-paramdef="%s"><textarea name="%s">%s</textarea></div>""")%(
            self.pd.name,
            self.pd.name,
            value
            )



# Time vartypes (keyformat is string)
#    Indexed as keyformat 'str'
import dateutil.parser
import dateutil.tz

@Vartype.register('datetime')
class vt_datetime(vt_string):
    """ISO 8601 Date time."""

    def validate(self, value):
        ret = []
        for i in ci(value):
            if i:
                t = dateutil.parser.parse(i)
                if not t.tzinfo:
                    raise ValidationError, "No UTC offset: %s"%i
                ret.append(t.isoformat())
        return self._rci(ret)
    
    def _unicode(self, value):
        tz = self.options.get('tz')
        raw_time = dateutil.parser.parse(value)
        local_time = raw_time.astimezone(dateutil.tz.gettz(tz))
        try:
            return local_time.strftime("%Y-%m-%d %H:%M")
        except:
            return "Date out of bounds! %s"%value
    
    def _html(self, value):
        tz = self.options.get('tz')
        raw_time = dateutil.parser.parse(value)
        raw_utc = raw_time.astimezone(dateutil.tz.gettz())
        local_time = raw_time.astimezone(dateutil.tz.gettz(tz))
        return Markup("""<time class="e2-edit" data-paramdef="%s" datetime="%s" title="Raw value: %s \nUTC time: %s \nLocal time: %s">%s</time>""")%(
            self.pd.name,
            raw_time.isoformat(),
            value,
            raw_utc.isoformat(),
            local_time.isoformat(),
            local_time.strftime("%Y-%m-%d %H:%M")
            )

@Vartype.register('date')
class vt_date(vt_datetime):
    """Date, yyyy-mm-dd."""
    def validate(self, value):
        ret = []
        for i in ci(value):
            i = parse_iso8601(i, result='date')
            if i:
                ret.append(i)
        return self._rci(ret)


@Vartype.register('time')
class vt_time(vt_datetime):
    """Time, HH:MM:SS."""
    def validate(self, value):
        ret = []
        for i in ci(value):
            i = parse_iso8601(i, result='time')
            if i:
                ret.append(i)
        return self._rci(ret)



# Reference vartypes.
#    Indexed as keyformat 'str'
@Vartype.register('uri')
class vt_uri(Vartype):
    """URI"""

    # ian: todo: parse with urlparse
    def validate(self, value):
        value = [unicode(i).strip() for i in ci(value)]
        for v in value:
            if not v.startswith("http://"):
                raise ValidationError, "Invalid URI: %s"%value
        return self._rci(value)



# Mapping types
@Vartype.register('dict')
class vt_dict(Vartype):
    """Dictionary with string keys and values."""

    keyformat = None

    def validate(self, value):
        if not value:
            return None
        r = [(unicode(k), unicode(v)) for k,v in value.items() if k]
        return dict(r)


@Vartype.register('dictlist')
class vt_dictlist(vt_dict):
    """Dictionary with string keys and list values."""

    keyformat = None

    def validate(self, value):
        if not value:
            return None        
        ret = {}
        for k,v in value.items():
            k = unicode(k)
            v = [unicode(i) for i in v]
            ret[k] = v
        return ret


# Binary vartypes
#    Not indexed.
@Vartype.register('binary')
class vt_binary(Vartype):
    """File Attachment"""

    def validate(self, value):
        value = self._validate_reference(ci(value), keytype='binary')
        return self._rci(value)

    def _unicode(self, value):
        try:
            v = self.db.binary.get(value)
            return v.filename or value
        except (ValueError, TypeError, AttributeError, KeyError), e:
            return 'Error getting binary: %s'%value

    def _add(self):
        label = "Change"
        multiple = ""
        if self.pd.iter:
            label = "Add attachments"
            multiple = "multiple"
        return Markup("""%s <input type="file" name="%s" multiple="%s" />""")%(
            label,
            self.pd.name,
            multiple
            )

    def _form(self, value):
        elem = ""
        if value:
            bdo = self.db.binary.get(value)
            src = "/download/%s/thumb.jpg?size=thumb"%(bdo.name)        
            elem = Markup("""
                <div class="e2-infobox" data-name="%s" data-keytype="user">
                    <input type="checkbox" name="%s" value="%s" checked="checked" />
                    <img src="%s" class="e2l-thumbnail" alt="Photo" />
                    <h4>%s</h4>
                    <p class="e2l-small">%s</p>                
                </div>""")%(
                    value,
                    self.pd.name,
                    value,
                    src,
                    bdo.filename,
                    bdo.filesize
                )
        
        # Show a 'Change' button...
        if not self.pd.iter:
            elem += self._add()
            
        return Markup("""<div class="e2-edit" %s>%s</div>""")%(
            self.pd.name,
            elem
            )



# md5 checksum
#    Indexed as keyformat 'str'
@Vartype.register('md5')
class vt_md5(Vartype):
    """String"""

    def validate(self, value):
        return self._rci([unicode(x).strip() for x in ci(value)])



# References to other database objects
#    Not indexed.
@Vartype.register('record')
class vt_record(Vartype):
    """References to other Records."""

    # This ma change in the future
    keyformat = 'str'

    def validate(self, value):
        value = self._validate_reference(ci(value), keytype='record')
        return self._rci(value)


@Vartype.register('link')
class vt_link(Vartype):
    """Reference."""
    def validate(self, value):
        value = self._validate_reference(ci(value), keytype='record') # Ugly hack :(
        return self._rci(value)



# User, ACL, and Group vartypes
#    Indexed as keyformat 'str'
@Vartype.register('user')
class vt_user(Vartype):
    """Users."""
    
    def validate(self, value):
        value = self._validate_reference(ci(value), keytype='user')
        return self._rci(value)

    def _unicode(self, value):
        update_user_cache(self.cache, self.db, [value])
        key = ('user', value)
        hit, user = self.cache.check(key)
        if user:
            return user.getdisplayname(lnf=self.options.get('lnf'))
        return value

    def _add(self):
        label = "Change"
        iter_ = ""
        if self.pd.iter:
            label = "+"
            iter_ = "true"
        return Markup("""<input type="button" value="%s" class="e2-edit-add-find" data-keytype="user" data-param="%s" data-iter="%s"/>""")%(
            label,
            self.pd.name,
            iter_
            )

    def _form(self, value):
        if value is None:
            return ""

        update_user_cache(self.cache, self.db, [value])
        key = ('user', value)
        hit, user = self.cache.check(key)

        src = "/static/images/user.png"
        if user.userrec.get('person_photo'):
            src = "/download/%s/user.jpg?size=thumb"%(user.userrec.get('person_photo'))

        elem = Markup("""
            <div class="e2-infobox" data-name="%s" data-keytype="user">
                <input type="checkbox" name="%s" value="%s" checked="checked" />
                <img src="%s" class="e2l-thumbnail" alt="Photo" />
                <h4>%s</h4>
                <p class="e2l-small">%s</p>                
            </div>
            """)%(
                value,
                self.pd.name,
                value,
                src,
                user.getdisplayname(),
                user.email
            )
        
        # Show a 'Change' button...
        if not self.pd.iter:
            elem += self._add()
            
        return Markup("""<div class="e2-edit" data-paramdef="%s">%s</div>""")%(
            self.pd.name,
            elem
            )
        

@Vartype.register('acl')
class vt_acl(Vartype):
    """Permissions access control list; nested lists of users."""
    
    def validate(self, value):
        if not hasattr(value, '__iter__'):
            value = [[value],[],[],[]]

        if hasattr(value, 'items'):
            v = [[],[],[],[]]
            ci = emen2.util.listops.check_iterable
            v[0] = ci(value.get('read'))
            v[1] = ci(value.get('comment'))
            v[2] = ci(value.get('write'))
            v[3] = ci(value.get('admin'))
            value = v

        for i in value:
            if not hasattr(i, '__iter__'):
                raise ValidationError, "Invalid permissions format: %s"%(value)

        value = [[unicode(y) for y in x] for x in value]
        if len(value) != 4:
            raise ValidationError, "Invalid permissions format: ", value

        users = reduce(lambda x,y:x+y, value)
        self._validate_reference(users, keytype='user')
        return value

    def reindex(self, items):
        """(Internal) Calculate secrindex updates"""
        # Calculating security updates...
        addrefs = collections.defaultdict(list)
        delrefs = collections.defaultdict(list)
        for name, new, old in items:
            nperms = set()
            for i in new or []:
                nperms |= set(i)
            operms = set()
            for i in old or []:
                operms |= set(i)

            for user in nperms - operms:
                addrefs[user].append(name)

            for user in operms - nperms:
                delrefs[user].append(name)

        return addrefs, delrefs

    def _form(self, value):
        return self._html(value)


@Vartype.register('group')
class vt_group(Vartype):
    """Group."""
    
    def validate(self, value):
        value = self._validate_reference(ci(value), keytype='group')
        return self._rci(value)


# Comment and History vartypes
#    Not indexed
@Vartype.register('comments')
class vt_comments(Vartype):
    """Comments."""
    
    keyformat = None

    # ian: todo... sort this out.
    def validate(self, value):
        return value
      
    def _html(self, value):
        user, dt, comment = value
        update_user_cache(self.cache, self.db, [user])
        key = ('user', user)
        hit, user = self.cache.check(key)
        return Markup("""%s said on <time class="e2-localize" datetime="%s">%s</time>: %s""")%(
                user.getdisplayname(),
                dt,
                dt,
                comment
            )

    def render_form(self, value):
        return Markup("""<div class="e2-edit" data-paramdef="%s"><textarea placeholder="Add additional comments"></textarea></div>""")%(self.pd.name)



@Vartype.register('history')
class vt_history(Vartype):
    """History."""
    
    keyformat = None




__version__ = "$Revision: 1.155 $".split(":")[1][:-1].strip()



# $Id: properties.py,v 1.30 2013/05/01 16:32:12 irees Exp $
# This Python file is encoding:UTF-8 Encoded
"""Properties -- types of physical properties (units)

Classes:
    Property
    prop_*: A number of included Properties

"""

import sys
import math
import re

# EMEN2 imports
import emen2.util.listops
import emen2.db.exceptions
import emen2.db.magnitude as mg


# Convenience
ci = emen2.util.listops.check_iterable
ValidationError = emen2.db.exceptions.ValidationError

# Unit synonyms
equivs = {
    'mm^2': 'mm mm',
    'cm^2': 'cm cm',
    'm^2': 'm m',
    'km^2': 'km km',
    'square feet': 'ft ft',
    'square meters': 'm m',
    'square kilometers': 'km km',
    'square miles': 'mile mile',

    'gallons': 'gallon',

    'kg/m^3': 'kg/m m m',
    'mol/m^3': 'mol/m m m',    
    'm^3': 'm m m',

    'Torr': 'torr',

    u'Ångstrom': u'Å',
    u'Ångstroms': u'Å',
    'Angstrom': u'Å',
    'Angstroms': u'Å',

    'nL': 'nl',
    'uL': 'ul',
    'mL': 'ml',
    'L': 'l',
    'kL': 'kl',
    'ML': 'Ml',    

    'degrees K': 'K',
    'degrees Kelvin': 'K',
    'Kelvin': 'K',
    
    'degrees C': 'degC',
    'Celcius': 'degC',
    'centigrade': 'degC',
    'degrees Celcius': 'degC',
    
    'degrees': 'degree',
    
    'moles': 'mol',
    
    'dollar': '$',
    'dollars': '$',

    'feet': 'ft',
    'foot': 'ft',
    
    'inches': 'inch',
    
    'lightspeed': 'c',

    'dots per inch': 'dpi',
    'lines per inch': 'lpi',
    
    'Henry': 'H',
    'Henrys': 'H',
    
    'katal': 'kat',
    'katals': 'kat',
    
    'lumen': 'lm',
    'lumens': 'lm',
        
    'minute': 'min',
    'minutes': 'min',
    
    'mins': 'min',
    'days': 'day',
    'years': 'year'
}


# SI prefix synonyms
si_equivs = {
    'meter': 'm',
    'meters': 'm',
    'metre': 'm',
    'metres': 'm',

    'Liter': 'l',
    'Liters': 'l',
    'Litre': 'l',
    'Litres': 'l',

    'gram': 'g',
    'grams': 'g',

    'amp': 'ampere',
    'amps': 'ampere',
    'amperes': 'ampere',
    
    'second': 's',
    'seconds': 's',

    'Newton': 'N',
    'Newtons': 'N',
    
    'Gray': 'Gy',
    'Grays': 'Gy',
    
    'Watt': 'W',
    'Watts': 'W',
    
    'Joule': 'J',
    'Joules': 'J',
    
    'Volt': 'V',
    'Volts': 'V',
    
    'rads': 'rad',
    'radians': 'rad',
        
    'Hertz': 'Hz',
    
    'Ohm': 'ohm',
    'Ohms': 'ohm',
    
    'Tesla': 'T',
    'Teslas': 'T',
    
    'Sievert': 'Sv',
    'Sieverts': 'Sv',
    
    'Becquerel': 'Bq',
    'Becquerels': 'Bq',

    'Coulomb': 'C',
    'Coulombs': 'C',
    
    'farad': 'F',
    'farads': 'F'    
}

# SI Prefixes
si_prefix = {
    'yocto': 'y',
    'zepto': 'z',
    'atto': 'a',
    'femto': 'f',
    'pico': 'p',
    'nano': 'n',
    'micro': 'u',
    'milli': 'm',
    'centi': 'c',
    'deci': 'd',
    'kilo': 'k',
    'mega': 'M',
    'giga': 'G',
    'tera': 'T',
    'peta': 'P',
    'exa': 'E',
    'zetta': 'Z',
    'yotta': 'Y',
}

# bit/byte synonyms
bytes_equivs = {
    'bit': 'b',
    'bits': 'b',
    'byte': 'B',
    'bytes': 'B',
}

# SI bit/byte prefixes
bytes_prefix = {
    'Kibi': 'Ki',
    'Mebi': 'Mi',
    'Gibi': 'Gi',
    'Tebi': 'Ti',
    'Pebi': 'Pi',
    'Exbi': 'Ei'
}

# Massage my equivalents into the units system.
for name,abbr in si_equivs.items():
    equivs[name] = abbr
    for prefix, p in si_prefix.items():
        equivs[prefix + name] = p + abbr

for name,abbr in bytes_equivs.items():
    equivs[name] = abbr
    for prefix, p in si_prefix.items():
        equivs[prefix + name] = p + abbr
    for prefix, p in bytes_prefix.items():
        equivs[prefix + name] = p + abbr
    
for name, abbr in equivs.items():
    equivs[name.lower()] = abbr

# Structural biology units: Angstrom and Dalton
mg.new_mag(u'Å', mg.Magnitude(1e-10, m=1))
mg.new_mag('Da', mg.Magnitude(1.6605402e-27, kg=1))

# Unitless
mg.new_mag('pixel', mg.Magnitude(1.0))
mg.new_mag('count', mg.Magnitude(1.0))
mg.new_mag('unitless', mg.Magnitude(1))
mg.new_mag('%RH', mg.Magnitude(1))
mg.new_mag('%RH', mg.Magnitude(1))
mg.new_mag('%T', mg.Magnitude(1))
mg.new_mag('%', mg.Magnitude(1))
mg.new_mag('pfu', mg.Magnitude(1))

# Non-SI Units
mg.new_mag('degF', mg.Magnitude(1.0))
mg.new_mag('mile', mg.mg(160934.4, 'cm'))
mg.new_mag('gallon', mg.Magnitude(3.78541178e-3, m=3))
mg.new_mag('torr', mg.Magnitude(1/760.0, m=-1, kg=1, s=-2))
mg.new_mag('degree', mg.Magnitude(1))


class Property(object):    
    restricted = False
    defaultunits = None
    units = []
    
    ##### Extensions #####

    registered = {}
    @classmethod
    def register(cls, name):
        def f(o):
            if name in cls.registered:
                raise ValueError("""%s is already registered""" % name)
            cls.registered[name] = o
            return o
        return f

    @classmethod
    def get_property(cls, name, *args, **kwargs):
        return cls.registered[name](*args, **kwargs)

    ##### Validation #####

    def validate(self, engine, pd, value, db):
        if hasattr(value, "__iter__"):
            return [self.validate(engine, pd, i, db) for i in value]

        if hasattr(value, "__float__"):
            return float(value)

        # print "Parsing for units: '%s'"%value            
        value = unicode(value).strip()
        if not value:
            return None

        # Match floating point numbers
        q = re.compile("^([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)(.*)")

        try:
            r = q.match(value).groups()
        except Exception, e:
            raise ValueError, "Unable to parse '%s' for units"%(value)

        value = float(r[0])
        
        target = pd.defaultunits
        if not target:
            target = self.defaultunits

        units = None
        if r[2] != None:
            units = unicode(r[2]).strip()
        if not units:
            units = target

        return self.convert(value, units, target)

    def check_bounds(self, value, target):
        return value

    def check_restrict(self, units, allowed):
        return
        # allowed = [equivs.get(i,i) for i in allowed]
        # if units not in allowed:
        #     raise ValueError, "Units %s not allowed for this property. Allowed units: %s"%(units, ", ".join(allowed))
            
    def convert(self, value, units, target):
        units = equivs.get(units, units)
        target = equivs.get(target, target)    
        if not units or units == target:
            self.check_bounds(value, target)
            return value

        # self.check_restrict(units, self.units)
        value = self._convert(value, units, target)
        self.check_bounds(value, target)
        return value

    def _convert(self, value, units, target):
        # print "value/units/target", value, units, target
        try:
            v = mg.mg(value, units, ounit=target)
        except emen2.db.magnitude.MagnitudeError:
            raise
            
        # Check for dimensionless conversions...
        if v.dimensionless():
            self.error(units, target, msg='dimensionless property')        
        return v.toval()

    def error(self, units, target, msg=None):
        raise ValueError, "Couldn't convert %s to %s: %s"%(units, target, msg or '')        

    def unknown(self, units, target):
        raise ValueError, "Don't know how to convert %s to %s"%(units, target)

@Property.register('transmittance')
class prop_transmittance(Property):
    defaultunits = '%T'
    units = ['%T']

@Property.register('force')
class prop_force(Property):
    defaultunits = 'N'
    units = ['N']

@Property.register('energy')
class prop_energy(Property):
    defaultunits = 'J'
    units = ['J']

@Property.register('resistance')
class prop_resistance(Property):
    defaultunits = 'ohm'
    units = ['microohm', 'milliohm', 'ohm']

@Property.register('dose')
class prop_dose(Property):
    defaultunits = "e/A2/sec"
    units = ["e/A2/sec"]

@Property.register('exposure')
class prop_exposure(Property):
    defaultunits = 'e/A^2'
    units = ['e/A^2']

@Property.register('currency')
class prop_currency(Property):
    defaultunits = 'dollar'
    units = ['dollar']

@Property.register('voltage')
class prop_voltage(Property):
    defaultunits = 'V'
    units = ['uV', 'mV', 'V', 'kV', 'MV']

@Property.register('pH')
class prop_pH(Property):
    defaultunits = 'pH'
    units = ['pH']
    def check_bounds(self, value, target):
        if (target == 'pH' and not 0 <- value <- 14):
            raise ValueError, "pH must be between 0 and 14"

@Property.register('concentration')
class prop_concentration(Property):
    defaultunits = 'mg/ml'
    units = ['mg/ml', 'pfu', 'kg/kg', 'kg/m^3', 'mol/m^3', 'mol/mol']

@Property.register('angle')
class prop_angle(Property):
    defaultunits = 'rad'
    units = ['mrad', 'rad', 'degree']

    def _convert(self, value, units, target):
        # if we get one of the rad variants, convert to rads
        if units != 'degree' and target == 'degree':
            value = mg.mg(value, units, ounit='rad')
            value = float(value) * (180.0 / math.pi)
            
        elif units == 'degree' and target != 'degree':
            value = value * (math.pi / 180.0)
            value = mg.mg(value, 'rad', target)
    
        elif units != 'degree' and target != 'degree':
            value = mg.mg(value, units, ounit=target)

        else:
            self.unknown(units, target)

        return value

@Property.register('temperature')
class prop_temperature(Property):
    defaultunits = 'K'
    units = ['degC', 'K', 'degF']
    
    def check_bounds(self, value, target):
        # Check absolute zero
        if (target == 'degC' and value < -273.15) or (target == 'K' and value < 0) or (target == 'degF' and value < -459.67):
            raise ValueError, "Cannot set a temperature below absolute zero"

    def _convert(self, value, units, target):
        # C / K
        if units == 'degC' and target == 'K':
            value = value + 273.15
        elif units == 'K' and target == 'degC':
            value = value - 273.15
        # F / K
        elif units == 'degF' and target == 'K':
            value = (value + 459.67) * (5.0 / 9.0)
        elif units == 'K' and target == 'degF':
            value = (value * (9.0 / 5.0)) - 459.67
        # C / F
        elif units == 'degC' and target == 'degF':
            value = value * (9.0 / 5.0) + 32
        elif units == 'degF' and target == 'degC':
            value = (value - 32) * (5.0 / 9.0)
        else:
            raise ValueError, "Don't know how to convert %s to %s"%(units, target)
        return value
    
@Property.register('area')
class prop_area(Property):
    defaultunits = 'm^2'
    units = ['mm^2', 'cm^2', 'm^2', 'km^2', 'square feet']

@Property.register('current')        
class prop_current(Property):
    defaultunits = 'A'
    units = ['uA', 'mA', 'A', 'kA', 'MA']

@Property.register('bytes')
class prop_bytes(Property):
    defaultunits = 'B'
    units = ['B', 'KB', 'MB', 'GB', 'TB', 'KiB', 'MiB', 'GiB', 'TiB']

@Property.register('percentage')
class prop_percentage(Property):
    defaultunits = '%'
    units = ['%']
    
@Property.register('property')
class prop_momentum(Property):
    defaultunits = 'kg m/s'
    units = ['kg m/s']

@Property.register('volume')
class prop_volume(Property):
    defaultunits = 'L'
    units = ['nL', 'uL', 'mL', 'L', 'gallon', 'm^3']

@Property.register('pressure')
class prop_pressure(Property):
    defaultunits = 'Pa'
    units = ['Pa', 'bar', 'atm', 'torr', 'psi']
    
@Property.register('unitless')
class prop_unitless(Property):
    defaultunits = 'unitless'
    units = ['unitless']
    
@Property.register('inductance')
class prop_inductance(Property):
    defaultunits = 'H'
    units = ['H']
        
@Property.register('currentdensity')
class prop_currentdensity(Property):
    defaultunits = 'Pi Amp/cm^2'
    units = ['Pi Amp/cm^2']
    
@Property.register('count')
class prop_count(Property):
    defaultunits = 'count'
    units = ['count', 'pixel']

@Property.register('bfactor')
class prop_bfactor(Property):
    defaultunits = 'A^2'
    units = ['A^2']

@Property.register('relative_humidity')
class prop_relative_humidity(Property):
    defaultunits = '%RH'
    units = ['%RH']

@Property.register('length')
class prop_length(Property):
    defaultunits = 'm'
    units = [u'Å', 'nm', 'um', 'mm', 'm', 'km']

@Property.register('mass')
class prop_mass(Property):
    defaultunits = 'g'
    units = ['Da', 'kDa', 'MDa', 'ng', 'ug', 'mg', 'g', 'kg']

@Property.register('time')
class prop_time(Property):
    defaultunits = 's'
    units = ['fs', 'ps', 'ns', 'us', 'ms', 's', 'min', 'hour', 'day', 'year']

@Property.register('velocity')
class prop_velocity(Property):
    defaultunits = 'm/s'
    units = ['m/s']

@Property.register('acceleration')
class prop_acceleration(Property):
    defaultunits = 'm/s**2'
    units = ['m/s**2']

@Property.register('resolution')
class prop_resolution(Property):
    defaultunits = u'Å/pixel'
    units = [u'Å/pixel', 'dpi', 'lpi']
        

if __name__ == '__main__':
    # print convert(-460, 'degF', 'K')
    # a = prop_temperature()
    # print a.convert(100, 'K', 'degC')
    
    # a = prop_angle()
    # print a.convert(360, 'degree', 'rad')    
        
    # a = prop_mass()
    # print a.convert(1, 'kDa', 'MDa')    
        
    a = prop_angle()
    print a.convert(1, 'degree', 'rads')    
        
__version__ = "$Revision: 1.30 $".split(":")[1][:-1].strip()


# $Id: responsecodes.py,v 1.9 2012/07/28 06:31:19 irees Exp $
'''Contains Classses which should mirror HTTP resoponse codes'''
#NOTE: unittests in tests/responsecodes_test.py -- run after any changes


class HTTPResponseCode(Exception):
    '''Base class for setting HTTP Response codes.  If a view
    raises one of these, the resource should catch it and set
    the response code appropriately'''
    title = None
    code = 500
    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg
        self.headers = {}


class MethodNotSupported(HTTPResponseCode):
    '''Set the response status to 501'''
    code = 501


class HTTPCancelledResponse(HTTPResponseCode):
    '''Set the response status to 500'''
    code = 500


class HTTP200Response(HTTPResponseCode):
    '''Set the response status to 200'''
    code = 200


class HTTP300Response(HTTPResponseCode):
    '''Set the response status to 300'''
    code = 300


class HTTPMovedPermanently(HTTP300Response):
    '''Set the response status to 301.

    If this is caught the resource should send a redirect'''
    code = 301
    def __init__(self, msg, dest):
        HTTP300Response.__init__(self, msg)
        self.headers['Location'] = unicode(dest).encode('utf-8')


class HTTPFound(HTTP300Response):
    '''Set the response status to 302

    If this is caught the resource should send a redirect'''
    code = 302
    def __init__(self, msg, dest):
        HTTP300Response.__init__(self, msg)
        self.headers['Location'] = unicode(dest).encode('utf-8')


class HTTPNotModified(HTTP300Response):
    '''Set the response status to 304'''
    code = 304


class HTTP400Response(HTTPResponseCode):
    '''Set the response status to 400'''
    code = 400


class UnauthorizedError(HTTP400Response):
    '''Set the response status to 401'''
    code = 401


class ForbiddenError(HTTP400Response):
    '''Set the response status to 403'''
    code = 403


class NotFoundError(HTTP400Response):
    '''Set the response status to 404 (Not Found)'''
    title = 'Page Not Found'
    msg = 'The requested URL (%s) was not found on this server.'
    code = 404
    def __init__(self, msg):
        self.msg %= msg
        HTTP400Response.__init__(self, self.msg)


class MethodNotAllowedError(HTTP400Response):
    '''Set the response status to 405'''
    title = 'Method Not Allowed'
    msg = 'Method Not Allowed: %r'
    code = 405
    def __init__(self, msg):
        self.msg %= msg
        HTTP400Response.__init__(self, self.msg)


class GoneError(HTTP400Response):
    '''Set the response status to 410'''
    code = 410



__version__ = "$Revision: 1.9 $".split(":")[1][:-1].strip()



# $Id: paramdef.py,v 1.27 2013/05/01 08:22:10 irees Exp $
import collections

from emen2.web.view import View


@View.register
class ParamDef(View):

    @View.add_matcher(r'^/paramdef/(?P<name>[^/]*)/$')
    def main(self, name=None):
        self.paramdef = self.db.get(name, filt=False, keytype='paramdef')
        self.template = '/pages/paramdef.main'
        self.title = "Parameter: %s"%self.paramdef.desc_short

        parentmap = self.routing.execute('Tree/embed', db=self.db, keytype='paramdef', root=self.paramdef.name, mode='parents', recurse=3)

        units = set()
        if self.paramdef and self.paramdef.property:
            units = self.db.paramdef.units(self.paramdef.property)

        vartypes = self.db.paramdef.vartypes()
        properties = self.db.paramdef.properties()

        self.ctxt.update(dict(
            paramdef = self.paramdef,
            create = self.db.auth.check.create(),
            editable = self.paramdef.writable(),
            vartypes = vartypes,
            properties = properties,
            edit = False,
            new = False,
            parentmap = parentmap
            ))


    @View.add_matcher(r'^/paramdef/(?P<name>[^/]*)/edit/$')
    def edit(self, name, **kwargs):
        if self.request_method == 'post':            
            paramdef = self.db.get(name, filt=False, keytype='paramdef')
            paramdef.update(kwargs)
            pd = self.db.paramdef.put(paramdef)
            if pd:
                self.redirect(self.routing.reverse('ParamDef/main', name=pd.name))
                return
    
        self.main(name=name)
        self.template = '/pages/paramdef.edit'
        self.ctxt['edit'] = True
        self.title = 'Edit Parameter: %s'%self.paramdef.desc_short
        
        
    @View.add_matcher(r'^/paramdef/(?P<name>[^/]*)/new/$')
    def new(self, name, **kwargs):
        if self.request_method == 'post':
            vartype = kwargs.pop('vartype', None)            
            paramdef = self.db.paramdef.new(name=name, vartype=vartype)
            paramdef.update(kwargs)
            pd = self.db.paramdef.put(paramdef)
            if pd:
                self.redirect(self.routing.reverse('ParamDef/main', name=pd.name))
                return

        self.main(name=name)
        self.template = '/pages/paramdef.new'
        self.paramdef.parents = set([self.paramdef.name])
        self.paramdef.children = set()
        self.ctxt['edit'] = True
        self.ctxt['new'] = True
        self.title = 'New Parameter based on: %s'%self.paramdef.desc_short


@View.register
class ParamDefs(View):

    @View.add_matcher(r'^/paramdefs/$')
    def main(self, action=None, q=None):
        paramdefnames = self.db.paramdef.filter(None)

        if action == None or action not in ["vartype", "name", "tree", "property"]:
            action = "tree"

        if q:
            action = "name"
            paramdefs = self.db.paramdef.find(q)
        else:
            paramdefs = self.db.get(paramdefnames, keytype='paramdef')

        # Tab Switcher
        pages = collections.OrderedDict()
        pages['tree'] = 'Parameter ontology'
        pages['name'] = 'Parameters by name'
        pages['vartype'] = 'Parameters by data type'
        pages['property'] = 'Parameters by property'
        uris = {}
        for k in pages:
            uris[k] = self.routing.reverse('ParamDefs/%s'%k)
        pages.uris = uris
        pages.active = action
        self.ctxt['pages'] = pages

        self.template = '/pages/paramdefs.%s'%action        
        self.title = pages.get(action)
        
        # Children
        childmap = self.routing.execute('Tree/embed', db=self.db, mode="children", keytype="paramdef", root="root", recurse=-1, id='sitemap')

        self.ctxt['paramdefnames'] = paramdefnames
        self.ctxt['paramdefs'] = paramdefs
        self.ctxt['q'] = q
        self.ctxt['childmap'] = childmap
        self.ctxt['create'] = self.db.auth.check.create()


    @View.add_matcher(r'^/paramdefs/vartype/$')
    def vartype(self, *args, **kwargs):
        return self.main(action='vartype', *args, **kwargs)


    @View.add_matcher(r'^/paramdefs/tree/$')
    def tree(self, *args, **kwargs):
        return self.main(action='tree', *args, **kwargs)


    @View.add_matcher(r'^/paramdefs/property/$')
    def property(self, *args, **kwargs):
        return self.main(action='property', *args, **kwargs)
        
        
    @View.add_matcher(r'^/paramdefs/name/$')
    def name(self, *args, **kwargs):
        return self.main(action='name', *args, **kwargs)
        






__version__ = "$Revision: 1.27 $".split(":")[1][:-1].strip()



# # $Id: eman2resource.py,v 1.4 2012/10/18 09:41:40 irees Exp $
# import re
# import os
# import pickle
# import traceback
# import time
# import random
# import cStringIO
# import tempfile
# import subprocess
# import atexit
# import signal
# 
# import jsonrpc.jsonutil
# 
# # Twisted Imports
# from twisted.internet import threads
# from twisted.python import failure, filepath
# from twisted.internet import threads
# from twisted.internet import defer
# from twisted.web.resource import Resource
# from twisted.web.static import *
# 
# import emen2.db.config
# 
# 
# 
# def loadEMAN():
#     # Try to import EMAN2..
#     try:
#         import EMAN2
#         # We need to steal these handlers from EMAN2...
#         # signal.signal(2, signal.SIG_DFL)
#         # signal.signal(15, signal.SIG_DFL)
#         # atexit.register(emen2.db.database.DB_Close)
#     except ImportError:
#         EMAN2 = None
#     return EMAN2
# 
# 
# 
# class RingDict(dict):
# 
#     def __init__(self, *args, **kwargs):
#         self.maxlen = kwargs.pop('maxlen', None)
#         self.order = [] # collections.deque()
#         dict.__init__(self, *args, **kwargs)
# 
# 
#     def __getitem__(self, key):
#         if key in self.order:
#             self.order.remove(key)
#         self.order.insert(0, key)
#         return dict.__getitem__(self, key)
# 
# 
#     def __delitem__(self, key):
#         if key in self.order:
#             self.order.remove(key)
#         dict.__delitem__(self, key)
# 
# 
#     def __setitem__(self, key, value):
#         if key in self.order:
#             self.order.remove(key)
# 
#         self.order.insert(0, key)
#         self.order = self.order[:self.maxlen]
# 
#         for i in self.keys():
#             if i not in self.order:
#                 del self[i]
# 
#         dict.__setitem__(self, key, value)
# 
# 
#     # def _check_compress(self, bdo):
#     #     compress = False
#     #     filepath = bdo.get('filepath')
#     #     workfile = bdo.get('filepath')
#     #
#     #     fs = bdo.get('filename').split(".")
#     #     if fs[-1] == "gz":
#     #         compress = "gzip"
#     #     elif fs[-1] == "bz2":
#     #         compress = "bzip2"
#     #
#     #     if compress:
#     #         fs.pop()
#     #         workfile = os.path.join(g.paths.tmp, "%s.%s"%(bdo.name.replace(":","."), fs[-1]))
#     #         if not os.access(workfile, os.F_OK):
#     #             print "Decompressing %s to %s"%(filepath, workfile)
#     #             a = subprocess.Popen("%s -d -c %s > %s"%(compress, filepath, workfile), shell=True)
#     #             a.wait()
#     #
#     #     return workfile
#     #
#     #
#     # def _open(self, bdo, ctxid=None):
#     #     key = bdo.name
#     #     emd = EMEN2RING.get(key)
#     #     if emd:
#     #         return emd
#     #
#     #     workfile = str(self._check_compress(bdo))
#     #
#     #     emd = EMAN2.EMData()
#     #     emd.read_image(workfile)
#     #     emd.process_inplace("normalize")
#     #     rmin = emd.get_attr("mean") - emd.get_attr("sigma") * 3.0
#     #     rmax = emd.get_attr("mean") + emd.get_attr("sigma") * 3.0
#     #     emd.set_attr("render_min", rmin)
#     #     emd.set_attr("render_max", rmax)
#     #     EMEN2RING[key] = emd
#     #     return emd
# 
# 
# 
# 
# ##########################################
# # Download Resource
# 
# 
# class EMAN2BoxResource(Resource, File):
#     """This class keeps an internal ring buffer of open EMData objects, and renders boxes cut out from them"""
# 
#     isLeaf = True
#     defaultType="application/octet-stream"
#     ring = RingDict(maxlen=10)
# 
#     def render(self, request):
#         host = request.getClientIP()
#         args = request.args
#         request.postpath = filter(bool, request.postpath)
#         ctxid = request.getCookie("ctxid")
# 
#         if request.args.get("ctxid"):
#             ctxid = request.args.get("ctxid",[None])[0]
# 
#         bdo = request.postpath[0]
#         method = request.postpath[1]
#         d = threads.deferToThread(self.action, bdo, method, request.args, ctxid, host)
#         d.addErrback(self._ebRender, request)
# 
#         if method == "box":
#             d.addCallback(self._boxRender, request)
#         else:
#             d.addCallback(self._cbRender, request)
# 
#         return server.NOT_DONE_YET
# 
# 
#     def _check_compress(self, bdo):
# 
#         compress = False
#         filepath = bdo.get('filepath')
#         workfile = bdo.get('filepath')
# 
#         fs = bdo.get('filename').split(".")
#         if fs[-1] == "gz":
#             compress = "gzip"
#         elif fs[-1] == "bz2":
#             compress = "bzip2"
# 
#         if compress:
#             fs.pop()
#             tmppath = emen2.db.config.get('paths.tmp')
#             workfile = os.path.join(tmppath, "%s.%s"%(bdo.name.replace(":","."), fs[-1]))
#             if not os.access(workfile, os.F_OK):
#                 # print "Decompressing %s to %s"%(filepath, workfile)
#                 a = subprocess.Popen("%s -d -c %s > %s"%(compress, filepath, workfile), shell=True)
#                 a.wait()
# 
#         return workfile
# 
# 
#     def _open(self, bdo, ctxid=None):
#         import EMAN2
# 
#         key = bdo.name
#         emd = self.ring.get(key)
#         if emd:
#             return emd
# 
#         workfile = str(self._check_compress(bdo))
# 
#         emd = EMAN2.EMData()
#         emd.read_image(workfile)
#         emd.process_inplace("normalize")
#         rmin = emd.get_attr("mean") - emd.get_attr("sigma") * 3.0
#         rmax = emd.get_attr("mean") + emd.get_attr("sigma") * 3.0
#         emd.set_attr("render_min", rmin)
#         emd.set_attr("render_max", rmax)
#         self.ring[key] = emd
#         return emd
# 
# 
# 
#     def method_auto_contrast(self, bdo, args):
#         emd = self._open(bdo)
#         rmin = emd.get_attr("mean") - emd.get_attr("sigma") * 3.0
#         rmax = emd.get_attr("mean") + emd.get_attr("sigma") * 3.0
#         emd.set_attr("render_min", rmin)
#         emd.set_attr("render_max", rmax)
#         return emd.get_attr_dict()
# 
# 
#     def method_get_attr_dict(self, bdo, args):
#         emd = self._open(bdo)
#         return emd.get_attr_dict()
# 
# 
#     def method_get_range(self, bdo, args):
#         emd = self._open(bdo)
#         return emd.get_attr("render_min"), emd.get_attr("render_max")
# 
# 
#     def method_set_range(self, bdo, args):
#         rmin = int(args.get("min", [0])[0])
#         rmax = int(args.get("max", [0])[0])
#         emd = self._open(bdo)
#         emd.set_attr("render_min", rmin)
#         emd.set_attr("render_max", rmax)
#         return emd.get_attr("render_min"), emd.get_attr("render_max")
# 
# 
# 
#     def method_autobox(self, bdo, args):
#         # particle_diameter=128
# 
#         emd = self._open(bdo)
#         boxertools.set_idd_image_entry(bdo.get(name), boxertools.FLCFImage.DB_NAME, emd)
# 
# 
# 
#         boxes = [[1024,1024], [512, 1024], [2048, 2048]]
#         boxer = e2boxer.SwarmBoxer()
#         boxer.handle_file_change(filename)
#         boxer_vitals = emboxerbase.EMBoxerModuleVitals(file_names=[filename])
#         boxer_vitals.current_idx=0
#         boxer.target = weakref.ref(boxer_vitals)
#         boxer.clear_all()
#         for box in boxes:
#             boxer.add_ref(box[0], box[1], filename)
# 
#         ret = boxer.auto_box(filename, True, True, cache=False)
#         return ret
# 
# 
# 
#     def method_box(self, bdo, args):
#         import EMAN2
# 
#         size = float(args.get("size",[128])[0])
#         scale = float(args.get("scale",[1])[0])
#         x = float(args.get("x", [0])[0])
#         y = float(args.get("y", [0])[0])
#         fill = args.get("fill", [None])[0]
#         emd = self._open(bdo)
# 
# 
#         region = EMAN2.Region(x, y, size, size)
#         if fill:
#             cutout = emd.get_clip(region, fill=emd.get_attr('maximum'))
#         else:
#             cutout = emd.get_clip(region)
# 
# 
#         # print "imgsize: %s size: %s scale: %s x: %s y: %s"%(emd.get_attr('nx'), size, scale, x, y)
# 
#         if scale > 1:
#             #cutout.process_inplace("math.meanshrink", {"n":scale})
#             cutout.process_inplace("xform.scale", {"scale":1/scale})
# 
#         # rmin, rmax = emd.get_attr("render_min"), emd.get_attr("render_max")
#         if args.has_key("min") or args.has_key("max"):
#             rmin = float(args.get("min", [0])[0])
#             rmax = float(args.get("max", [0])[0])
#             cutout.set_attr("render_min", rmin)
#             cutout.set_attr("render_max", rmax)
# 
#         outfile = tempfile.NamedTemporaryFile(delete=False, suffix=".png")
# 
#         f = outfile.name
#         outfile.close()
#         cutout.write_image(f)
#         return f
# 
# 
#     def action(self, bdo, method, args, ctxid, host, db=None):
# 
#         with db:
#             db._setContext(ctxid,host)
#             bdo = db.binary.get(bdo)
# 
#         if not bdo:
#             raise IOError
# 
#         if method == "box":
#             return self.method_box(bdo, args)
# 
#         elif method == "set_range":
#             return self.method_set_range(bdo, args)
# 
#         elif method == "get_range":
#             return self.method_get_range(bdo, args)
# 
#         elif method == "get_attr_dict":
#             return self.method_get_attr_dict(bdo, args)
# 
#         elif method == "auto_contrast":
#             return self.method_auto_contrast(bdo, args)
# 
# 
# 
#     def cbRender(self, result, request):
#         result = jsonrpc.jsonutil.encode(result).encode("utf-8")
#         request.write(result)
#         request.finish()
# 
# 
#     def _boxRender(self, result, request):
#         """You know what you doing."""
#         request.setHeader('content-type', 'image/png')
#         fsize = size = os.stat(result).st_size
#         f = open(result)
#         FileTransfer(f, size, request)
#         # f.close()
# 
# 
#     def dbRender(self,failure,request):
# 
#         errmsg = "Unspecified Error"
#         errcode = 500
# 
#         try:
#             failure.raiseException()
#         except IOError,e:
#             errcode = 404
#             errmsg = "File Not Found"
#         except Exception,e:
#             errmsg = str(e)
#             raise
# 
#         data = errmsg
# 
#         request.setResponseCode(errcode)
#         request.setHeader("content-type", "text/html; charset=utf-8")
#         request.setHeader('content-length',len(data))
#         request.write(data)
# 
#         request.finish()
# 
# 
# 
# __version__ = "$Revision: 1.4 $".split(":")[1][:-1].strip()



# $Id: em.py,v 1.47 2013/05/14 09:34:57 irees Exp $
import datetime
import time
import tempfile
import os

import twisted.web.static

import emen2.db.exceptions
import emen2.db.config
import emen2.db.log
from emen2.web.view import View




@View.register
class EMEquipment(View):
    @View.add_matcher(r'^/em/equipment/(?P<name>[^/]*)/$')
    def main(self, name, **kwargs):
        self.title = 'Equipment'
        self.template = '/em/project.main'

    @View.add_matcher(r'^/em/equipment/new/(?P<rectype>[^/]*)/$')
    def new(self, rectype, **kwargs):
        self.title = 'New Equipment'
        self.template = '/em/project.new'
        


@View.register
class EMHome(View):

    @View.add_matcher(r'^/$', view='Root', name='main')
    @View.add_matcher(r'^/em/home/$', r'^/db/home/$')
    def main(self, hideinactive=False, sortkey='name', reverse=False):
        self.title = 'Home'
        self.template = '/em/home.main'

        banner = None
        render_banner = ''
        
        if not self.ctxt['USER']:            
            self.template = '/em/home.noauth'
            try:
                banner = self.db.record.get(emen2.db.config.get('bookmarks.banner_noauth'))
                render_banner = self.db.view(banner, viewname="banner")
            except:
                pass
            self.ctxt['banner'] = banner
            self.ctxt['render_banner'] = render_banner
            return

        try:
            banner = self.db.record.get(emen2.db.config.get('bookmarks.banner'))
            render_banner = self.db.view(banner, viewname="banner")
        except:
            pass
            
        self.ctxt['banner'] = banner
        self.ctxt['render_banner'] = render_banner

        # Recent records
        now = datetime.datetime.utcnow().isoformat()+'+00:00'
        since = (datetime.datetime.utcnow() - datetime.timedelta(days=90)).isoformat()+'+00:00'
        q = self.db.plot(
            [['creationtime', '>=', since]], 
            x={'key':'creationtime', 'bin':'day', 'min':since, 'max':now}, 
            y={'stacked':True},
            sortkey='creationtime'
            )            
        self.ctxt['recent_activity'] = q

        # Table
        q_table = self.routing.execute('Query/embed', db=self.db, q={'count':20, 'subset':q['names']}, controls=False)
        self.ctxt['recent_activity_table'] = q_table
        
        
        # Groups and projects
        torender = set()
        def nodeleted(items):
            return filter(lambda x:not x.get('deleted'), items)

        # Groups
        groups = nodeleted(self.db.record.get(self.db.record.findbyrectype('group')))
        groupnames = set([i.name for i in groups])
        torender |= groupnames

        # Top-level children of groups (any rectype)
        groups_children = self.db.rel.children(groupnames)
        projs = set()
        for v in groups_children.values():
            projs |= v
        torender |= projs

        # Get projects, most recent children, and progress reports
        # projects_children = self.db.rel.children(projs, recurse=-1)
        projects_children = {}

        # Get all the recent records we want to display
        recnames = self.db.view(torender)

        # Update context
        self.ctxt['recnames'] = recnames
        self.ctxt['groups'] = groups
        self.ctxt['groups_children'] = groups_children
        self.ctxt['projects_children'] = {}
        
                
        
@View.register
class EMAN2Convert(View):
    contentTypes = twisted.web.static.loadMimeTypes()
    contentEncodings = {
            ".gz" : "gzip",
            ".bz2": "bzip2"
            }

    defaultType = 'application/octet-stream'    
    return_file = None
    
    @View.add_matcher(r'^/eman2/(?P<name>.+)/convert/(?P<format>\w+)/$', r'^/eman2/(?P<name>.+)/convert/$')
    def convert(self, name, format, normalize=False):
        import EMAN2

        if format not in ['tif', 'tiff', 'tif8', 'mrc', 'hdf', 'jpg', 'jpeg', 'png']:
            raise ValueError, "Invalid format: %s"%format

        depth = None
        if format in ['tif8']:
            depth = 8
            format = 'tif'

        bdo = self.db.binary.get(name)
        img = EMAN2.EMData()
        img.read_image(str(bdo.filepath))
        
        if normalize:
            img.process_inplace("normalize")            
        
        outfile = tempfile.NamedTemporaryFile(delete=False, suffix='.%s'%format)

        if depth == 8:
            img['render_min'] = -1
            img['render_max'] = 256
            img.write_image(str(outfile.name), -1, EMAN2.EMUtil.ImageType.IMAGE_UNKNOWN, False, None, EMAN2.EMUtil.EMDataType.EM_UCHAR, False)
        else:            
            img.write_image(str(outfile.name))

        filename = os.path.splitext(bdo.filename)[0]
        filename = '%s.%s'%(filename, format)
        return filename, outfile.name

    def render_result(self, result, request, t=0, **_):
        filename, filepath = result
        mimetype, encoding = twisted.web.static.getTypeAndEncoding(filename, self.contentTypes, self.contentEncodings, self.defaultType)

        fsize = os.stat(filepath).st_size
        f = open(filepath)

        request.setHeader('Content-Disposition', 'attachment; filename=%s'%filename.encode('utf-8'))
        request.setHeader('Content-Length', str(fsize))
        request.setHeader('Content-Type', mimetype)
        request.setHeader('Content-Encoding', encoding)

        a = twisted.web.static.NoRangeStaticProducer(request, f)
        a.start()

        try:
            emen2.db.log.info("Removing temporary file: %s"%filepath)
            os.remove(filepath)
        except:
            emen2.db.log.error("Couldn't remove temporary file: %s"%filepath)
            



# $Id: error.py,v 1.7 2012/09/14 06:26:24 irees Exp $
from emen2.web.view import View

@View.register
class Error(View):

    @View.add_matcher('/error/')
    @View.provides('error_handler')
    def main(self, error='', location='/', **kwargs):
        self.template = '/errors/error'
        self.title = 'Error'
        self.ctxt["error"] = error
        self.ctxt['location'] = location

    @View.add_matcher('/error/auth')
    def auth(self, error='', location='/', **kwargs):
        self.template = '/errors/auth'
        self.title = 'Error'
        self.ctxt["error"] = error
        self.ctxt['location'] = location

    @View.add_matcher('/error/resp')
    def resp(self, error='', location='/', **kwargs):
        self.template = '/errors/resp'
        self.title = 'Error'
        self.ctxt["error"] = error
        self.ctxt['location'] = location


        
__version__ = "$Revision: 1.7 $".split(":")[1][:-1].strip()



# $Id: rpc.py,v 1.1 2012/09/17 22:00:10 irees Exp $

if __name__ == "__main__":
    # Start the web server directly
    import emen2.web.server
    emen2.web.server.start_rpc()

__version__ = "$Revision: 1.1 $".split(":")[1][:-1].strip()
